@@ -16,7 +16,7 @@ function main({ n, statSyncType }) {
   const fn = fs[statSyncType];
 
   bench.start();
-  for (var i = 0; i < n; i++) {
+  for (let i = 0; i < n; i++) {
     fn(arg);
   }
   bench.end(n);
