@@ -14,12 +14,13 @@ const bench = common.createBenchmark(main, {
   type: ['asc', 'utf', 'buf'],
   len: [64 * 1024, 128 * 1024, 256 * 1024, 1024 * 1024],
   c: [100],
-  method: ['write', 'end']
+  method: ['write', 'end'],
+  duration: 5
 });
 
-function main({ len, type, method, c }) {
+function main({ len, type, method, c, duration }) {
   const http = require('http');
-  var chunk;
+  let chunk;
   switch (type) {
     case 'buf':
       chunk = Buffer.alloc(len, 'x');
@@ -43,14 +44,15 @@ function main({ len, type, method, c }) {
 
   const fn = method === 'write' ? write : end;
 
-  const server = http.createServer(function(req, res) {
+  const server = http.createServer((req, res) => {
     fn(res);
   });
 
-  server.listen(common.PORT, function() {
+  server.listen(common.PORT, () => {
     bench.http({
-      connections: c
-    }, function() {
+      connections: c,
+      duration
+    }, () => {
       server.close();
     });
   });
