@@ -9,15 +9,15 @@ const bench = common.createBenchmark(main, {
     'C:\\..\\',
     'C:\\foo',
     'C:\\foo\\bar',
-    'C:\\foo\\bar\\\\baz\\asdf\\quux\\..'
+    'C:\\foo\\bar\\\\baz\\asdf\\quux\\..',
   ],
-  n: [1e6]
+  n: [1e5]
 });
 
 function main({ n, path }) {
   bench.start();
-  for (var i = 0; i < n; i++) {
-    win32.normalize(path);
+  for (let i = 0; i < n; i++) {
+    win32.normalize(i % 3 === 0 ? `${path}${i}` : path);
   }
   bench.end(n);
 }
