@@ -14,8 +14,8 @@ function main({ n }) {
   s._read = noop;
 
   bench.start();
-  for (var k = 0; k < n; ++k) {
-    for (var i = 0; i < 1e4; ++i)
+  for (let k = 0; k < n; ++k) {
+    for (let i = 0; i < 1e4; ++i)
       s.push(b);
     while (s.read());
   }
