@@ -1,7 +1,7 @@
 'use strict';
 const common = require('../common.js');
 const url = require('url');
-const hrefs = require('../fixtures/url-inputs.js').urls;
+const hrefs = common.urls;
 hrefs.noscheme = 'some.ran/dom/url.thing?oh=yes#whoo';
 
 const paths = {
@@ -23,7 +23,7 @@ function main({ n, href, path }) {
   const p = paths[path];
 
   bench.start();
-  for (var i = 0; i < n; i += 1)
+  for (let i = 0; i < n; i += 1)
     url.resolve(h, p);
   bench.end(n);
 }
