@@ -5,7 +5,7 @@
 #ifndef V8_OBJECTS_PROPERTY_DESCRIPTOR_OBJECT_INL_H_
 #define V8_OBJECTS_PROPERTY_DESCRIPTOR_OBJECT_INL_H_
 
-#include "src/objects-inl.h"
+#include "src/objects/objects-inl.h"
 #include "src/objects/property-descriptor-object.h"
 
 // Has to be the last include (doesn't have include guards):
@@ -14,6 +14,7 @@
 namespace v8 {
 namespace internal {
 
+OBJECT_CONSTRUCTORS_IMPL(PropertyDescriptorObject, FixedArray)
 CAST_ACCESSOR(PropertyDescriptorObject)
 }
 }  // namespace v8
