@@ -15,12 +15,12 @@
 #include <vector>
 
 #include "include/v8-profiler.h"
-#include "src/allocation.h"
 #include "src/base/platform/time.h"
 #include "src/builtins/builtins.h"
-#include "src/code-events.h"
+#include "src/codegen/source-position.h"
+#include "src/logging/code-events.h"
 #include "src/profiler/strings-storage.h"
-#include "src/source-position.h"
+#include "src/utils/allocation.h"
 
 namespace v8 {
 namespace internal {
@@ -234,7 +234,36 @@ struct CodeEntryAndLineNumber {
   int line_number;
 };
 
-typedef std::vector<CodeEntryAndLineNumber> ProfileStackTrace;
+struct ProfileStackFrame {
+  CodeEntryAndLineNumber entry;
+  Address native_context;
+  bool filterable;  // If true, the frame should be filtered by context (if a
+                    // filter is present).
+};
+
+typedef std::vector<ProfileStackFrame> ProfileStackTrace;
+
+// Filters stack frames from sources other than a target native context.
+class ContextFilter {
+ public:
+  explicit ContextFilter(Address native_context_address)
+      : native_context_address_(native_context_address) {}
+
+  // Returns true if the stack frame passes a context check.
+  bool Accept(const ProfileStackFrame&);
+
+  // Invoked when a native context has changed address.
+  void OnMoveEvent(Address from_address, Address to_address);
+
+  // Update the context's tracked address based on VM-thread events.
+  void set_native_context_address(Address address) {
+    native_context_address_ = address;
+  }
+  Address native_context_address() const { return native_context_address_; }
+
+ private:
+  Address native_context_address_;
+};
 
 class ProfileTree;
 
@@ -311,7 +340,7 @@ class V8_EXPORT_PRIVATE ProfileTree {
   explicit ProfileTree(Isolate* isolate);
   ~ProfileTree();
 
-  typedef v8::CpuProfilingMode ProfilingMode;
+  using ProfilingMode = v8::CpuProfilingMode;
 
   ProfileNode* AddPathFromEnd(
       const std::vector<CodeEntry*>& path,
@@ -321,7 +350,8 @@ class V8_EXPORT_PRIVATE ProfileTree {
       const ProfileStackTrace& path,
       int src_line = v8::CpuProfileNode::kNoLineNumberInfo,
       bool update_stats = true,
-      ProfilingMode mode = ProfilingMode::kLeafNodeLineNumbers);
+      ProfilingMode mode = ProfilingMode::kLeafNodeLineNumbers,
+      ContextFilter* context_filter = nullptr);
   ProfileNode* root() const { return root_; }
   unsigned next_node_id() { return next_node_id_++; }
   unsigned GetFunctionId(const ProfileNode* node);
@@ -358,20 +388,22 @@ class CpuProfiler;
 
 class CpuProfile {
  public:
-  typedef v8::CpuProfilingMode ProfilingMode;
-
   struct SampleInfo {
     ProfileNode* node;
     base::TimeTicks timestamp;
     int line;
   };
 
-  CpuProfile(CpuProfiler* profiler, const char* title, bool record_samples,
-             ProfilingMode mode);
+  V8_EXPORT_PRIVATE CpuProfile(CpuProfiler* profiler, const char* title,
+                               CpuProfilingOptions options);
 
+  // Checks whether or not the given TickSample should be (sub)sampled, given
+  // the sampling interval of the profiler that recorded it (in microseconds).
+  V8_EXPORT_PRIVATE bool CheckSubsample(base::TimeDelta sampling_interval);
   // Add pc -> ... -> main() call path to the profile.
   void AddPath(base::TimeTicks timestamp, const ProfileStackTrace& path,
-               int src_line, bool update_stats);
+               int src_line, bool update_stats,
+               base::TimeDelta sampling_interval);
   void FinishProfile();
 
   const char* title() const { return title_; }
@@ -380,9 +412,14 @@ class CpuProfile {
   int samples_count() const { return static_cast<int>(samples_.size()); }
   const SampleInfo& sample(int index) const { return samples_[index]; }
 
+  int64_t sampling_interval_us() const {
+    return options_.sampling_interval_us();
+  }
+
   base::TimeTicks start_time() const { return start_time_; }
   base::TimeTicks end_time() const { return end_time_; }
   CpuProfiler* cpu_profiler() const { return profiler_; }
+  ContextFilter* context_filter() const { return context_filter_.get(); }
 
   void UpdateTicksScale();
 
@@ -392,8 +429,8 @@ class CpuProfile {
   void StreamPendingTraceEvents();
 
   const char* title_;
-  bool record_samples_;
-  ProfilingMode mode_;
+  const CpuProfilingOptions options_;
+  std::unique_ptr<ContextFilter> context_filter_;
   base::TimeTicks start_time_;
   base::TimeTicks end_time_;
   std::deque<SampleInfo> samples_;
@@ -401,6 +438,9 @@ class CpuProfile {
   CpuProfiler* const profiler_;
   size_t streaming_next_sample_;
   uint32_t id_;
+  // Number of microseconds worth of profiler ticks that should elapse before
+  // the next sample is recorded.
+  base::TimeDelta next_sample_delta_;
 
   static std::atomic<uint32_t> last_id_;
 
@@ -447,11 +487,9 @@ class V8_EXPORT_PRIVATE CpuProfilesCollection {
  public:
   explicit CpuProfilesCollection(Isolate* isolate);
 
-  typedef v8::CpuProfilingMode ProfilingMode;
-
   void set_cpu_profiler(CpuProfiler* profiler) { profiler_ = profiler; }
-  bool StartProfiling(const char* title, bool record_samples,
-                      ProfilingMode mode = ProfilingMode::kLeafNodeLineNumbers);
+  bool StartProfiling(const char* title, CpuProfilingOptions options = {});
+
   CpuProfile* StopProfiling(const char* title);
   std::vector<std::unique_ptr<CpuProfile>>* profiles() {
     return &finished_profiles_;
@@ -460,10 +498,19 @@ class V8_EXPORT_PRIVATE CpuProfilesCollection {
   bool IsLastProfile(const char* title);
   void RemoveProfile(CpuProfile* profile);
 
+  // Finds a common sampling interval dividing each CpuProfile's interval,
+  // rounded up to the nearest multiple of the CpuProfiler's sampling interval.
+  // Returns 0 if no profiles are attached.
+  base::TimeDelta GetCommonSamplingInterval() const;
+
   // Called from profile generator thread.
   void AddPathToCurrentProfiles(base::TimeTicks timestamp,
                                 const ProfileStackTrace& path, int src_line,
-                                bool update_stats);
+                                bool update_stats,
+                                base::TimeDelta sampling_interval);
+
+  // Called from profile generator thread.
+  void UpdateNativeContextAddressForCurrentProfiles(Address from, Address to);
 
   // Limits the number of profiles that can be simultaneously collected.
   static const int kMaxSimultaneousProfiles = 100;
@@ -482,18 +529,20 @@ class V8_EXPORT_PRIVATE CpuProfilesCollection {
 
 class V8_EXPORT_PRIVATE ProfileGenerator {
  public:
-  explicit ProfileGenerator(CpuProfilesCollection* profiles);
+  explicit ProfileGenerator(CpuProfilesCollection* profiles, CodeMap* code_map);
 
   void RecordTickSample(const TickSample& sample);
 
-  CodeMap* code_map() { return &code_map_; }
+  void UpdateNativeContextAddress(Address from, Address to);
+
+  CodeMap* code_map() { return code_map_; }
 
  private:
   CodeEntry* FindEntry(Address address);
   CodeEntry* EntryForVMState(StateTag tag);
 
   CpuProfilesCollection* profiles_;
-  CodeMap code_map_;
+  CodeMap* const code_map_;
 
   DISALLOW_COPY_AND_ASSIGN(ProfileGenerator);
 };
