@@ -21,7 +21,12 @@
 
 'use strict';
 
-const { Object } = primordials;
+const {
+  Array,
+  ArrayIsArray,
+  ObjectDefineProperty,
+  ObjectSetPrototypeOf,
+} = primordials;
 
 const errors = require('internal/errors');
 const {
@@ -30,15 +35,11 @@ const {
   newHandle,
 } = require('internal/dgram');
 const { guessHandleType } = internalBinding('util');
-const {
-  isLegalPort,
-} = require('internal/net');
 const {
   ERR_INVALID_ARG_TYPE,
   ERR_MISSING_ARGS,
   ERR_SOCKET_ALREADY_BOUND,
   ERR_SOCKET_BAD_BUFFER_SIZE,
-  ERR_SOCKET_BAD_PORT,
   ERR_SOCKET_BUFFER_SIZE,
   ERR_SOCKET_CANNOT_SEND,
   ERR_SOCKET_DGRAM_IS_CONNECTED,
@@ -49,11 +50,12 @@ const {
 const {
   isInt32,
   validateString,
-  validateNumber
+  validateNumber,
+  validatePort,
 } = require('internal/validators');
 const { Buffer } = require('buffer');
 const { deprecate } = require('internal/util');
-const { isUint8Array } = require('internal/util/types');
+const { isArrayBufferView } = require('internal/util/types');
 const EventEmitter = require('events');
 const {
   defaultTriggerAsyncIdScope,
@@ -79,7 +81,7 @@ const RECV_BUFFER = true;
 const SEND_BUFFER = false;
 
 // Lazily loaded
-var cluster = null;
+let cluster = null;
 
 const errnoException = errors.errnoException;
 const exceptionWithHostPort = errors.exceptionWithHostPort;
@@ -87,12 +89,13 @@ const exceptionWithHostPort = errors.exceptionWithHostPort;
 
 function Socket(type, listener) {
   EventEmitter.call(this);
-  var lookup;
+  let lookup;
   let recvBufferSize;
   let sendBufferSize;
 
+  let options;
   if (type !== null && typeof type === 'object') {
-    var options = type;
+    options = type;
     type = options.type;
     lookup = options.lookup;
     recvBufferSize = options.recvBufferSize;
@@ -120,8 +123,8 @@ function Socket(type, listener) {
     sendBufferSize
   };
 }
-Object.setPrototypeOf(Socket.prototype, EventEmitter.prototype);
-Object.setPrototypeOf(Socket, EventEmitter);
+ObjectSetPrototypeOf(Socket.prototype, EventEmitter.prototype);
+ObjectSetPrototypeOf(Socket, EventEmitter);
 
 
 function createSocket(type, listener) {
@@ -210,10 +213,25 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
 
   state.bindState = BIND_STATE_BINDING;
 
-  if (arguments.length && typeof arguments[arguments.length - 1] === 'function')
-    this.once('listening', arguments[arguments.length - 1]);
+  const cb = arguments.length && arguments[arguments.length - 1];
+  if (typeof cb === 'function') {
+    function removeListeners() {
+      this.removeListener('error', removeListeners);
+      this.removeListener('listening', onListening);
+    }
 
-  if (port instanceof UDP) {
+    function onListening() {
+      removeListeners.call(this);
+      cb.call(this);
+    }
+
+    this.on('error', removeListeners);
+    this.on('listening', onListening);
+  }
+
+  if (port !== null &&
+      typeof port === 'object' &&
+      typeof port.recvStart === 'function') {
     replaceHandle(this, port);
     startListening(this);
     return this;
@@ -239,8 +257,8 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
       }, (err) => {
         // Callback to handle error.
         const ex = errnoException(err, 'open');
-        this.emit('error', ex);
         state.bindState = BIND_STATE_UNBOUND;
+        this.emit('error', ex);
       });
       return this;
     }
@@ -261,8 +279,8 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
     return this;
   }
 
-  var address;
-  var exclusive;
+  let address;
+  let exclusive;
 
   if (port !== null && typeof port === 'object') {
     address = port.address || '';
@@ -292,7 +310,7 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
     if (!cluster)
       cluster = require('cluster');
 
-    var flags = 0;
+    let flags = 0;
     if (state.reuseAddr)
       flags |= UV_UDP_REUSEADDR;
     if (state.ipv6Only)
@@ -308,8 +326,8 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
       }, (err) => {
         // Callback to handle error.
         const ex = exceptionWithHostPort(err, 'bind', ip, port);
-        this.emit('error', ex);
         state.bindState = BIND_STATE_UNBOUND;
+        this.emit('error', ex);
       });
     } else {
       if (!state.handle)
@@ -317,9 +335,9 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
 
       const err = state.handle.bind(ip, port || 0, flags);
       if (err) {
-        var ex = exceptionWithHostPort(err, 'bind', ip, port);
-        this.emit('error', ex);
+        const ex = exceptionWithHostPort(err, 'bind', ip, port);
         state.bindState = BIND_STATE_UNBOUND;
+        this.emit('error', ex);
         // Todo: close?
         return;
       }
@@ -331,21 +349,8 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
   return this;
 };
 
-
-function validatePort(port) {
-  const legal = isLegalPort(port);
-  if (legal)
-    port = port | 0;
-
-  if (!legal || port === 0)
-    throw new ERR_SOCKET_BAD_PORT(port);
-
-  return port;
-}
-
-
 Socket.prototype.connect = function(port, address, callback) {
-  port = validatePort(port);
+  port = validatePort(port, 'Port', { allowZero: false });
   if (typeof address === 'function') {
     callback = address;
     address = '';
@@ -451,29 +456,33 @@ Socket.prototype.sendto = function(buffer,
 function sliceBuffer(buffer, offset, length) {
   if (typeof buffer === 'string') {
     buffer = Buffer.from(buffer);
-  } else if (!isUint8Array(buffer)) {
+  } else if (!isArrayBufferView(buffer)) {
     throw new ERR_INVALID_ARG_TYPE('buffer',
-                                   ['Buffer', 'Uint8Array', 'string'], buffer);
+                                   ['Buffer',
+                                    'TypedArray',
+                                    'DataView',
+                                    'string'],
+                                   buffer);
   }
 
   offset = offset >>> 0;
   length = length >>> 0;
 
-  return buffer.slice(offset, offset + length);
+  return Buffer.from(buffer.buffer, buffer.byteOffset + offset, length);
 }
 
 
 function fixBufferList(list) {
   const newlist = new Array(list.length);
 
-  for (var i = 0, l = list.length; i < l; i++) {
-    var buf = list[i];
+  for (let i = 0, l = list.length; i < l; i++) {
+    const buf = list[i];
     if (typeof buf === 'string')
       newlist[i] = Buffer.from(buf);
-    else if (!isUint8Array(buf))
+    else if (!isArrayBufferView(buf))
       return null;
     else
-      newlist[i] = buf;
+      newlist[i] = Buffer.from(buf.buffer, buf.byteOffset, buf.byteLength);
   }
 
   return newlist;
@@ -513,8 +522,8 @@ function clearQueue() {
   state.queue = undefined;
 
   // Flush the send queue.
-  for (var i = 0; i < queue.length; i++)
-    queue[i]();
+  for (const queueEntry of queue)
+    queueEntry();
 }
 
 function isConnected(self) {
@@ -560,35 +569,45 @@ Socket.prototype.send = function(buffer,
       port = offset;
       address = length;
     }
-  } else if (typeof length === 'number') {
-    buffer = sliceBuffer(buffer, offset, length);
-    if (typeof port === 'function') {
-      callback = port;
-      port = null;
-    } else if (port || address) {
-      throw new ERR_SOCKET_DGRAM_IS_CONNECTED();
-    }
   } else {
-    callback = offset;
+    if (typeof length === 'number') {
+      buffer = sliceBuffer(buffer, offset, length);
+      if (typeof port === 'function') {
+        callback = port;
+        port = null;
+      }
+    } else {
+      callback = offset;
+    }
+
+    if (port || address)
+      throw new ERR_SOCKET_DGRAM_IS_CONNECTED();
   }
 
-  if (!Array.isArray(buffer)) {
+  if (!ArrayIsArray(buffer)) {
     if (typeof buffer === 'string') {
       list = [ Buffer.from(buffer) ];
-    } else if (!isUint8Array(buffer)) {
+    } else if (!isArrayBufferView(buffer)) {
       throw new ERR_INVALID_ARG_TYPE('buffer',
-                                     ['Buffer', 'Uint8Array', 'string'],
+                                     ['Buffer',
+                                      'TypedArray',
+                                      'DataView',
+                                      'string'],
                                      buffer);
     } else {
       list = [ buffer ];
     }
   } else if (!(list = fixBufferList(buffer))) {
     throw new ERR_INVALID_ARG_TYPE('buffer list arguments',
-                                   ['Buffer', 'string'], buffer);
+                                   ['Buffer',
+                                    'TypedArray',
+                                    'DataView',
+                                    'string'],
+                                   buffer);
   }
 
   if (!connected)
-    port = validatePort(port);
+    port = validatePort(port, 'Port', { allowZero: false });
 
   // Normalize callback so it's either a function or undefined but not anything
   // else.
@@ -662,6 +681,14 @@ function doSend(ex, self, ip, list, address, port, callback) {
   else
     err = state.handle.send(req, list, list.length, !!callback);
 
+  if (err >= 1) {
+    // Synchronous finish. The return code is msg_length + 1 so that we can
+    // distinguish between synchronous success and asynchronous success.
+    if (callback)
+      process.nextTick(callback, null, err - 1);
+    return;
+  }
+
   if (err && callback) {
     // Don't emit as error, dgram_legacy.js compatibility
     const ex = exceptionWithHostPort(err, 'send', address, port);
@@ -728,8 +755,8 @@ Socket.prototype.remoteAddress = function() {
   if (state.connectState !== CONNECT_STATE_CONNECTED)
     throw new ERR_SOCKET_DGRAM_NOT_CONNECTED();
 
-  var out = {};
-  var err = state.handle.getpeername(out);
+  const out = {};
+  const err = state.handle.getpeername(out);
   if (err)
     throw errnoException(err, 'getpeername');
 
@@ -820,6 +847,51 @@ Socket.prototype.dropMembership = function(multicastAddress,
   }
 };
 
+Socket.prototype.addSourceSpecificMembership = function(sourceAddress,
+                                                        groupAddress,
+                                                        interfaceAddress) {
+  healthCheck(this);
+
+  if (typeof sourceAddress !== 'string') {
+    throw new ERR_INVALID_ARG_TYPE('sourceAddress', 'string', sourceAddress);
+  }
+
+  if (typeof groupAddress !== 'string') {
+    throw new ERR_INVALID_ARG_TYPE('groupAddress', 'string', groupAddress);
+  }
+
+  const err =
+    this[kStateSymbol].handle.addSourceSpecificMembership(sourceAddress,
+                                                          groupAddress,
+                                                          interfaceAddress);
+  if (err) {
+    throw errnoException(err, 'addSourceSpecificMembership');
+  }
+};
+
+
+Socket.prototype.dropSourceSpecificMembership = function(sourceAddress,
+                                                         groupAddress,
+                                                         interfaceAddress) {
+  healthCheck(this);
+
+  if (typeof sourceAddress !== 'string') {
+    throw new ERR_INVALID_ARG_TYPE('sourceAddress', 'string', sourceAddress);
+  }
+
+  if (typeof groupAddress !== 'string') {
+    throw new ERR_INVALID_ARG_TYPE('groupAddress', 'string', groupAddress);
+  }
+
+  const err =
+    this[kStateSymbol].handle.dropSourceSpecificMembership(sourceAddress,
+                                                           groupAddress,
+                                                           interfaceAddress);
+  if (err) {
+    throw errnoException(err, 'dropSourceSpecificMembership');
+  }
+};
+
 
 function healthCheck(socket) {
   if (!socket[kStateSymbol].handle) {
@@ -891,7 +963,7 @@ Socket.prototype.getSendBufferSize = function() {
 
 
 // Deprecated private APIs.
-Object.defineProperty(Socket.prototype, '_handle', {
+ObjectDefineProperty(Socket.prototype, '_handle', {
   get: deprecate(function() {
     return this[kStateSymbol].handle;
   }, 'Socket.prototype._handle is deprecated', 'DEP0112'),
@@ -901,7 +973,7 @@ Object.defineProperty(Socket.prototype, '_handle', {
 });
 
 
-Object.defineProperty(Socket.prototype, '_receiving', {
+ObjectDefineProperty(Socket.prototype, '_receiving', {
   get: deprecate(function() {
     return this[kStateSymbol].receiving;
   }, 'Socket.prototype._receiving is deprecated', 'DEP0112'),
@@ -911,7 +983,7 @@ Object.defineProperty(Socket.prototype, '_receiving', {
 });
 
 
-Object.defineProperty(Socket.prototype, '_bindState', {
+ObjectDefineProperty(Socket.prototype, '_bindState', {
   get: deprecate(function() {
     return this[kStateSymbol].bindState;
   }, 'Socket.prototype._bindState is deprecated', 'DEP0112'),
@@ -921,7 +993,7 @@ Object.defineProperty(Socket.prototype, '_bindState', {
 });
 
 
-Object.defineProperty(Socket.prototype, '_queue', {
+ObjectDefineProperty(Socket.prototype, '_queue', {
   get: deprecate(function() {
     return this[kStateSymbol].queue;
   }, 'Socket.prototype._queue is deprecated', 'DEP0112'),
@@ -931,7 +1003,7 @@ Object.defineProperty(Socket.prototype, '_queue', {
 });
 
 
-Object.defineProperty(Socket.prototype, '_reuseAddr', {
+ObjectDefineProperty(Socket.prototype, '_reuseAddr', {
   get: deprecate(function() {
     return this[kStateSymbol].reuseAddr;
   }, 'Socket.prototype._reuseAddr is deprecated', 'DEP0112'),
@@ -953,7 +1025,7 @@ Socket.prototype._stopReceiving = deprecate(function() {
 
 // Legacy alias on the C++ wrapper object. This is not public API, so we may
 // want to runtime-deprecate it at some point. There's no hurry, though.
-Object.defineProperty(UDP.prototype, 'owner', {
+ObjectDefineProperty(UDP.prototype, 'owner', {
   get() { return this[owner_symbol]; },
   set(v) { return this[owner_symbol] = v; }
 });
