@@ -26,7 +26,15 @@
 // No new pull requests targeting this module will be accepted
 // unless they address existing, critical bugs.
 
-const util = require('util');
+const {
+  Array,
+  Error,
+  Map,
+  ObjectDefineProperty,
+  ReflectApply,
+  Symbol,
+} = primordials;
+
 const EventEmitter = require('events');
 const {
   ERR_DOMAIN_CALLBACK_NOT_AVAILABLE,
@@ -34,6 +42,7 @@ const {
   ERR_UNHANDLED_ERROR
 } = require('internal/errors').codes;
 const { createHook } = require('async_hooks');
+const { useDomainTrampoline } = require('internal/async_hooks');
 
 // TODO(addaleax): Use a non-internal solution for this.
 const kWeak = Symbol('kWeak');
@@ -41,8 +50,8 @@ const { WeakReference } = internalBinding('util');
 
 // Overwrite process.domain with a getter/setter that will allow for more
 // effective optimizations
-var _domain = [null];
-Object.defineProperty(process, 'domain', {
+const _domain = [null];
+ObjectDefineProperty(process, 'domain', {
   enumerable: true,
   get: function() {
     return _domain[0];
@@ -58,28 +67,31 @@ const asyncHook = createHook({
     if (process.domain !== null && process.domain !== undefined) {
       // If this operation is created while in a domain, let's mark it
       pairing.set(asyncId, process.domain[kWeak]);
-      resource.domain = process.domain;
-      if (resource.promise !== undefined &&
-          resource.promise instanceof Promise) {
-        // resource.promise instanceof Promise make sure that the
-        // promise comes from the same context
-        // see https://github.com/nodejs/node/issues/15673
-        resource.promise.domain = process.domain;
-      }
+      ObjectDefineProperty(resource, 'domain', {
+        configurable: true,
+        enumerable: false,
+        value: process.domain,
+        writable: true
+      });
     }
   },
   before(asyncId) {
     const current = pairing.get(asyncId);
-    if (current !== undefined) { // enter domain for this cb
+    if (current !== undefined) { // Enter domain for this cb
       // We will get the domain through current.get(), because the resource
       // object's .domain property makes sure it is not garbage collected.
+      // However, we do need to make the reference to the domain non-weak,
+      // so that it cannot be garbage collected before the after() hook.
+      current.incRef();
       current.get().enter();
     }
   },
   after(asyncId) {
     const current = pairing.get(asyncId);
-    if (current !== undefined) { // exit domain for this cb
-      current.get().exit();
+    if (current !== undefined) { // Exit domain for this cb
+      const domain = current.get();
+      current.decRef();
+      domain.exit();
     }
   },
   destroy(asyncId) {
@@ -123,7 +135,7 @@ function topLevelDomainCallback(cb, ...args) {
 
   if (domain)
     domain.enter();
-  const ret = Reflect.apply(cb, this, args);
+  const ret = ReflectApply(cb, this, args);
   if (domain)
     domain.exit();
 
@@ -134,7 +146,7 @@ function topLevelDomainCallback(cb, ...args) {
 // another one. The stack is each entered domain.
 const stack = [];
 exports._stack = stack;
-internalBinding('domain').enable(topLevelDomainCallback);
+useDomainTrampoline(topLevelDomainCallback);
 
 function updateExceptionCapture() {
   if (stack.every((domain) => domain.listenerCount('error') === 0)) {
@@ -194,16 +206,21 @@ exports.create = exports.createDomain = function createDomain() {
   return new Domain();
 };
 
-// the active domain is always the one that we're currently in.
+// The active domain is always the one that we're currently in.
 exports.active = null;
 Domain.prototype.members = undefined;
 
 // Called by process._fatalException in case an error was thrown.
 Domain.prototype._errorHandler = function(er) {
-  var caught = false;
+  let caught = false;
 
-  if (!util.isPrimitive(er)) {
-    er.domain = this;
+  if ((typeof er === 'object' && er !== null) || typeof er === 'function') {
+    ObjectDefineProperty(er, 'domain', {
+      configurable: true,
+      enumerable: false,
+      value: this,
+      writable: true
+    });
     er.domainThrown = true;
   }
 
@@ -233,7 +250,7 @@ Domain.prototype._errorHandler = function(er) {
       }
     }
   } else {
-    // wrap this in a try/catch so we don't get infinite throwing
+    // Wrap this in a try/catch so we don't get infinite throwing
     try {
       // One of three things will happen here.
       //
@@ -273,7 +290,7 @@ Domain.prototype._errorHandler = function(er) {
 
 
 Domain.prototype.enter = function() {
-  // note that this might be a no-op, but we still need
+  // Note that this might be a no-op, but we still need
   // to push it onto the stack so that we can pop it later.
   exports.active = process.domain = this;
   stack.push(this);
@@ -282,11 +299,11 @@ Domain.prototype.enter = function() {
 
 
 Domain.prototype.exit = function() {
-  // don't do anything if this domain is not on the stack.
-  var index = stack.lastIndexOf(this);
+  // Don't do anything if this domain is not on the stack.
+  const index = stack.lastIndexOf(this);
   if (index === -1) return;
 
-  // exit all domains until this one.
+  // Exit all domains until this one.
   stack.splice(index);
 
   exports.active = stack[stack.length - 1];
@@ -301,11 +318,11 @@ Domain.prototype.add = function(ee) {
   if (ee.domain === this)
     return;
 
-  // has a domain already - remove it first.
+  // Has a domain already - remove it first.
   if (ee.domain)
     ee.domain.remove(ee);
 
-  // check for circular Domain->Domain links.
+  // Check for circular Domain->Domain links.
   // This causes bad insanity!
   //
   // For example:
@@ -315,33 +332,38 @@ Domain.prototype.add = function(ee) {
   // e.add(d);
   // e.emit('error', er); // RangeError, stack overflow!
   if (this.domain && (ee instanceof Domain)) {
-    for (var d = this.domain; d; d = d.domain) {
+    for (let d = this.domain; d; d = d.domain) {
       if (ee === d) return;
     }
   }
 
-  ee.domain = this;
+  ObjectDefineProperty(ee, 'domain', {
+    configurable: true,
+    enumerable: false,
+    value: this,
+    writable: true
+  });
   this.members.push(ee);
 };
 
 
 Domain.prototype.remove = function(ee) {
   ee.domain = null;
-  var index = this.members.indexOf(ee);
+  const index = this.members.indexOf(ee);
   if (index !== -1)
     this.members.splice(index, 1);
 };
 
 
 Domain.prototype.run = function(fn) {
-  var ret;
+  let ret;
 
   this.enter();
   if (arguments.length >= 2) {
-    var len = arguments.length;
-    var args = new Array(len - 1);
+    const len = arguments.length;
+    const args = new Array(len - 1);
 
-    for (var i = 1; i < len; i++)
+    for (let i = 1; i < len; i++)
       args[i - 1] = arguments[i];
 
     ret = fn.apply(this, args);
@@ -356,22 +378,25 @@ Domain.prototype.run = function(fn) {
 
 function intercepted(_this, self, cb, fnargs) {
   if (fnargs[0] && fnargs[0] instanceof Error) {
-    var er = fnargs[0];
-    util._extend(er, {
-      domainBound: cb,
-      domainThrown: false,
-      domain: self
+    const er = fnargs[0];
+    er.domainBound = cb;
+    er.domainThrown = false;
+    ObjectDefineProperty(er, 'domain', {
+      configurable: true,
+      enumerable: false,
+      value: self,
+      writable: true
     });
     self.emit('error', er);
     return;
   }
 
-  var args = [];
-  var i, ret;
+  const args = [];
+  let ret;
 
   self.enter();
   if (fnargs.length > 1) {
-    for (i = 1; i < fnargs.length; i++)
+    for (let i = 1; i < fnargs.length; i++)
       args.push(fnargs[i]);
     ret = cb.apply(_this, args);
   } else {
@@ -384,7 +409,7 @@ function intercepted(_this, self, cb, fnargs) {
 
 
 Domain.prototype.intercept = function(cb) {
-  var self = this;
+  const self = this;
 
   function runIntercepted() {
     return intercepted(this, self, cb, arguments);
@@ -395,7 +420,7 @@ Domain.prototype.intercept = function(cb) {
 
 
 function bound(_this, self, cb, fnargs) {
-  var ret;
+  let ret;
 
   self.enter();
   if (fnargs.length > 0)
@@ -409,13 +434,18 @@ function bound(_this, self, cb, fnargs) {
 
 
 Domain.prototype.bind = function(cb) {
-  var self = this;
+  const self = this;
 
   function runBound() {
     return bound(this, self, cb, arguments);
   }
 
-  runBound.domain = this;
+  ObjectDefineProperty(runBound, 'domain', {
+    configurable: true,
+    enumerable: false,
+    value: this,
+    writable: true
+  });
 
   return runBound;
 };
@@ -425,7 +455,12 @@ EventEmitter.usingDomains = true;
 
 const eventInit = EventEmitter.init;
 EventEmitter.init = function() {
-  this.domain = null;
+  ObjectDefineProperty(this, 'domain', {
+    configurable: true,
+    enumerable: false,
+    value: null,
+    writable: true
+  });
   if (exports.active && !(this instanceof exports.Domain)) {
     this.domain = exports.active;
   }
@@ -445,7 +480,7 @@ EventEmitter.prototype.emit = function(...args) {
   // handler, there's no active domain or this is process
   if (shouldEmitError || domain === null || domain === undefined ||
       this === process) {
-    return Reflect.apply(eventEmit, this, args);
+    return ReflectApply(eventEmit, this, args);
   }
 
   if (type === 'error') {
@@ -454,7 +489,12 @@ EventEmitter.prototype.emit = function(...args) {
 
     if (typeof er === 'object') {
       er.domainEmitter = this;
-      er.domain = domain;
+      ObjectDefineProperty(er, 'domain', {
+        configurable: true,
+        enumerable: false,
+        value: domain,
+        writable: true
+      });
       er.domainThrown = false;
     }
 
@@ -463,7 +503,7 @@ EventEmitter.prototype.emit = function(...args) {
   }
 
   domain.enter();
-  const ret = Reflect.apply(eventEmit, this, args);
+  const ret = ReflectApply(eventEmit, this, args);
   domain.exit();
 
   return ret;
