@@ -1,5 +1,13 @@
 'use strict';
 
+const {
+  ArrayIsArray,
+  ObjectDefineProperty,
+  ObjectSetPrototypeOf,
+  Symbol,
+  Uint8Array,
+} = primordials;
+
 const {
   errnoException,
   codes: {
@@ -18,11 +26,17 @@ const { validateString } = require('internal/validators');
 const EventEmitter = require('events');
 const net = require('net');
 const dgram = require('dgram');
-const util = require('util');
-const assert = require('assert');
+const inspect = require('internal/util/inspect').inspect;
+const assert = require('internal/assert');
 
 const { Process } = internalBinding('process_wrap');
-const { WriteWrap } = internalBinding('stream_wrap');
+const {
+  WriteWrap,
+  kReadBytesOrError,
+  kArrayBufferOffset,
+  kLastWriteWasAsync,
+  streamBaseState
+} = internalBinding('stream_wrap');
 const { Pipe, constants: PipeConstants } = internalBinding('pipe_wrap');
 const { TCP } = internalBinding('tcp_wrap');
 const { TTY } = internalBinding('tty_wrap');
@@ -43,30 +57,29 @@ const {
   UV_ENOENT,
   UV_ENOSYS,
   UV_ESRCH
-} = process.binding('uv');
+} = internalBinding('uv');
 
 const { SocketListSend, SocketListReceive } = SocketList;
 
-// Lazy loaded for startup performance.
-let StringDecoder;
 // Lazy loaded for startup performance and to allow monkey patching of
 // internalBinding('http_parser').HTTPParser.
 let freeParser;
 let HTTPParser;
 
 const MAX_HANDLE_RETRANSMISSIONS = 3;
+const kIsUsedAsStdio = Symbol('kIsUsedAsStdio');
 
-// this object contain function to convert TCP objects to native handle objects
+// This object contain function to convert TCP objects to native handle objects
 // and back again.
 const handleConversion = {
   'net.Native': {
     simultaneousAccepts: true,
 
-    send: function(message, handle, options) {
+    send(message, handle, options) {
       return handle;
     },
 
-    got: function(message, handle, emit) {
+    got(message, handle, emit) {
       emit(handle);
     }
   },
@@ -74,32 +87,32 @@ const handleConversion = {
   'net.Server': {
     simultaneousAccepts: true,
 
-    send: function(message, server, options) {
+    send(message, server, options) {
       return server._handle;
     },
 
-    got: function(message, handle, emit) {
-      var server = new net.Server();
-      server.listen(handle, function() {
+    got(message, handle, emit) {
+      const server = new net.Server();
+      server.listen(handle, () => {
         emit(server);
       });
     }
   },
 
   'net.Socket': {
-    send: function(message, socket, options) {
+    send(message, socket, options) {
       if (!socket._handle)
         return;
 
-      // if the socket was created by net.Server
+      // If the socket was created by net.Server
       if (socket.server) {
-        // the worker should keep track of the socket
+        // The worker should keep track of the socket
         message.key = socket.server._connectionKey;
 
-        var firstTime = !this.channel.sockets.send[message.key];
-        var socketList = getSocketList('send', this, message.key);
+        const firstTime = !this.channel.sockets.send[message.key];
+        const socketList = getSocketList('send', this, message.key);
 
-        // the server should no longer expose a .connection property
+        // The server should no longer expose a .connection property
         // and when asked to close it should query the socket status from
         // the workers
         if (firstTime) socket.server._setupWorker(socketList);
@@ -109,9 +122,9 @@ const handleConversion = {
           socket.server._connections--;
       }
 
-      var handle = socket._handle;
+      const handle = socket._handle;
 
-      // remove handle from socket object, it will be closed when the socket
+      // Remove handle from socket object, it will be closed when the socket
       // will be sent
       if (!options.keepOpen) {
         handle.onread = nop;
@@ -121,7 +134,7 @@ const handleConversion = {
         if (freeParser === undefined)
           freeParser = require('_http_common').freeParser;
         if (HTTPParser === undefined)
-          HTTPParser = internalBinding('http_parser').HTTPParser;
+          HTTPParser = require('_http_common').HTTPParser;
 
         // In case of an HTTP connection socket, release the associated
         // resources
@@ -135,7 +148,7 @@ const handleConversion = {
       return handle;
     },
 
-    postSend: function(message, handle, options, callback, target) {
+    postSend(message, handle, options, callback, target) {
       // Store the handle after successfully sending it, so it can be closed
       // when the NODE_HANDLE_ACK is received. If the handle could not be sent,
       // just close it.
@@ -153,18 +166,18 @@ const handleConversion = {
       }
     },
 
-    got: function(message, handle, emit) {
-      var socket = new net.Socket({
+    got(message, handle, emit) {
+      const socket = new net.Socket({
         handle: handle,
         readable: true,
         writable: true
       });
 
-      // if the socket was created by net.Server we will track the socket
+      // If the socket was created by net.Server we will track the socket
       if (message.key) {
 
-        // add socket to connections list
-        var socketList = getSocketList('got', this, message.key);
+        // Add socket to connections list
+        const socketList = getSocketList('got', this, message.key);
         socketList.add({
           socket: socket
         });
@@ -177,11 +190,11 @@ const handleConversion = {
   'dgram.Native': {
     simultaneousAccepts: false,
 
-    send: function(message, handle, options) {
+    send(message, handle, options) {
       return handle;
     },
 
-    got: function(message, handle, emit) {
+    got(message, handle, emit) {
       emit(handle);
     }
   },
@@ -189,22 +202,37 @@ const handleConversion = {
   'dgram.Socket': {
     simultaneousAccepts: false,
 
-    send: function(message, socket, options) {
+    send(message, socket, options) {
       message.dgramType = socket.type;
 
       return socket[kStateSymbol].handle;
     },
 
-    got: function(message, handle, emit) {
-      var socket = new dgram.Socket(message.dgramType);
+    got(message, handle, emit) {
+      const socket = new dgram.Socket(message.dgramType);
 
-      socket.bind(handle, function() {
+      socket.bind(handle, () => {
         emit(socket);
       });
     }
   }
 };
 
+function stdioStringToArray(stdio, channel) {
+  const options = [];
+
+  switch (stdio) {
+    case 'ignore':
+    case 'pipe': options.push(stdio, stdio, stdio); break;
+    case 'inherit': options.push(0, 1, 2); break;
+    default:
+      throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
+  }
+
+  if (channel) options.push(channel);
+
+  return options;
+}
 
 function ChildProcess() {
   EventEmitter.call(this);
@@ -236,7 +264,7 @@ function ChildProcess() {
     this._handle = null;
 
     if (exitCode < 0) {
-      var syscall = this.spawnfile ? 'spawn ' + this.spawnfile : 'spawn';
+      const syscall = this.spawnfile ? 'spawn ' + this.spawnfile : 'spawn';
       const err = errnoException(exitCode, syscall);
 
       if (this.spawnfile)
@@ -248,7 +276,7 @@ function ChildProcess() {
       this.emit('exit', this.exitCode, this.signalCode);
     }
 
-    // if any of the stdio streams have not been touched,
+    // If any of the stdio streams have not been touched,
     // then pull all the data through so that it can get the
     // eof and emit a 'close' event.
     // Do it on nextTick so that the user has one last chance
@@ -259,7 +287,8 @@ function ChildProcess() {
     maybeClose(this);
   };
 }
-util.inherits(ChildProcess, EventEmitter);
+ObjectSetPrototypeOf(ChildProcess.prototype, EventEmitter.prototype);
+ObjectSetPrototypeOf(ChildProcess, EventEmitter);
 
 
 function flushStdio(subprocess) {
@@ -267,10 +296,16 @@ function flushStdio(subprocess) {
 
   if (stdio == null) return;
 
-  for (var i = 0; i < stdio.length; i++) {
+  for (let i = 0; i < stdio.length; i++) {
     const stream = stdio[i];
-    if (!stream || !stream.readable || stream._readableState.readableListening)
+    // TODO(addaleax): This doesn't necessarily account for all the ways in
+    // which data can be read from a stream, e.g. being consumed on the
+    // native layer directly as a StreamBase.
+    if (!stream || !stream.readable ||
+        stream._readableState.readableListening ||
+        stream[kIsUsedAsStdio]) {
       continue;
+    }
     stream.resume();
   }
 }
@@ -297,47 +332,55 @@ function closePendingHandle(target) {
 
 
 ChildProcess.prototype.spawn = function(options) {
-  var ipc;
-  var ipcFd;
-  var i;
+  let i = 0;
 
   if (options === null || typeof options !== 'object') {
     throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
   }
 
   // If no `stdio` option was given - use default
-  var stdio = options.stdio || 'pipe';
+  let stdio = options.stdio || 'pipe';
 
-  stdio = _validateStdio(stdio, false);
+  stdio = getValidStdio(stdio, false);
 
-  ipc = stdio.ipc;
-  ipcFd = stdio.ipcFd;
+  const ipc = stdio.ipc;
+  const ipcFd = stdio.ipcFd;
   stdio = options.stdio = stdio.stdio;
 
+  if (options.serialization !== undefined &&
+      options.serialization !== 'json' &&
+      options.serialization !== 'advanced') {
+    throw new ERR_INVALID_OPT_VALUE('options.serialization',
+                                    options.serialization);
+  }
+
+  const serialization = options.serialization || 'json';
+
   if (ipc !== undefined) {
     // Let child process know about opened IPC channel
     if (options.envPairs === undefined)
       options.envPairs = [];
-    else if (!Array.isArray(options.envPairs)) {
+    else if (!ArrayIsArray(options.envPairs)) {
       throw new ERR_INVALID_ARG_TYPE('options.envPairs',
                                      'Array',
                                      options.envPairs);
     }
 
-    options.envPairs.push('NODE_CHANNEL_FD=' + ipcFd);
+    options.envPairs.push(`NODE_CHANNEL_FD=${ipcFd}`);
+    options.envPairs.push(`NODE_CHANNEL_SERIALIZATION_MODE=${serialization}`);
   }
 
   validateString(options.file, 'options.file');
   this.spawnfile = options.file;
 
-  if (Array.isArray(options.args))
+  if (ArrayIsArray(options.args))
     this.spawnargs = options.args;
   else if (options.args === undefined)
     this.spawnargs = [];
   else
     throw new ERR_INVALID_ARG_TYPE('options.args', 'Array', options.args);
 
-  var err = this._handle.spawn(options);
+  const err = this._handle.spawn(options);
 
   // Run-time errors should emit an error, not throw an exception.
   if (err === UV_EACCES ||
@@ -346,12 +389,11 @@ ChildProcess.prototype.spawn = function(options) {
       err === UV_ENFILE ||
       err === UV_ENOENT) {
     process.nextTick(onErrorNT, this, err);
+
     // There is no point in continuing when we've hit EMFILE or ENFILE
     // because we won't be able to set up the stdio file descriptors.
-    // It's kind of silly that the de facto spec for ENOENT (the test suite)
-    // mandates that stdio _is_ set up, even if there is no process on the
-    // receiving end, but it is what it is.
-    if (err !== UV_ENOENT) return err;
+    if (err === UV_EMFILE || err === UV_ENFILE)
+      return err;
   } else if (err) {
     // Close all opened fds on error
     for (i = 0; i < stdio.length; i++) {
@@ -377,9 +419,22 @@ ChildProcess.prototype.spawn = function(options) {
       continue;
     }
 
+    // The stream is already cloned and piped, thus stop its readable side,
+    // otherwise we might attempt to read from the stream when at the same time
+    // the child process does.
+    if (stream.type === 'wrap') {
+      stream.handle.reading = false;
+      stream.handle.readStop();
+      stream._stdio.pause();
+      stream._stdio.readableFlowing = false;
+      stream._stdio._readableState.reading = false;
+      stream._stdio[kIsUsedAsStdio] = true;
+      continue;
+    }
+
     if (stream.handle) {
-      // when i === 0 - we're dealing with stdin
-      // (which is the only one writable pipe)
+      // When i === 0 - we're dealing with stdin
+      // (which is the only one writable pipe).
       stream.socket = createSocket(this.pid !== 0 ?
         stream.handle : null, i > 0);
 
@@ -405,7 +460,7 @@ ChildProcess.prototype.spawn = function(options) {
     this.stdio.push(stdio[i].socket === undefined ? null : stdio[i].socket);
 
   // Add .send() method and start listening for IPC data
-  if (ipc !== undefined) setupChannel(this, ipc);
+  if (ipc !== undefined) setupChannel(this, ipc, serialization);
 
   return err;
 };
@@ -422,7 +477,7 @@ ChildProcess.prototype.kill = function(sig) {
     convertToValidSignal(sig === undefined ? 'SIGTERM' : sig);
 
   if (this._handle) {
-    var err = this._handle.kill(signal);
+    const err = this._handle.kill(signal);
     if (err === 0) {
       /* Success. */
       this.killed = true;
@@ -472,11 +527,12 @@ class Control extends EventEmitter {
   }
 }
 
-function setupChannel(target, channel) {
+let serialization;
+function setupChannel(target, channel, serializationMode) {
   target.channel = channel;
 
   // _channel can be deprecated in version 8
-  Object.defineProperty(target, '_channel', {
+  ObjectDefineProperty(target, '_channel', {
     get() { return target.channel; },
     set(val) { target.channel = val; },
     enumerable: true
@@ -487,36 +543,28 @@ function setupChannel(target, channel) {
 
   const control = new Control(channel);
 
-  if (StringDecoder === undefined)
-    StringDecoder = require('string_decoder').StringDecoder;
-  var decoder = new StringDecoder('utf8');
-  var jsonBuffer = '';
-  var pendingHandle = null;
-  channel.buffering = false;
+  if (serialization === undefined)
+    serialization = require('internal/child_process/serialization');
+  const {
+    initMessageChannel,
+    parseChannelMessages,
+    writeChannelMessage
+  } = serialization[serializationMode];
+
+  let pendingHandle = null;
+  initMessageChannel(channel);
   channel.pendingHandle = null;
-  channel.onread = function(nread, pool) {
+  channel.onread = function(arrayBuffer) {
     const recvHandle = channel.pendingHandle;
     channel.pendingHandle = null;
-    // TODO(bnoordhuis) Check that nread > 0.
-    if (pool) {
+    if (arrayBuffer) {
+      const nread = streamBaseState[kReadBytesOrError];
+      const offset = streamBaseState[kArrayBufferOffset];
+      const pool = new Uint8Array(arrayBuffer, offset, nread);
       if (recvHandle)
         pendingHandle = recvHandle;
 
-      // Linebreak is used as a message end sign
-      var chunks = decoder.write(pool).split('\n');
-      var numCompleteChunks = chunks.length - 1;
-      // Last line does not have trailing linebreak
-      var incompleteChunk = chunks[numCompleteChunks];
-      if (numCompleteChunks === 0) {
-        jsonBuffer += incompleteChunk;
-        this.buffering = jsonBuffer.length !== 0;
-        return;
-      }
-      chunks[0] = jsonBuffer + chunks[0];
-
-      for (var i = 0; i < numCompleteChunks; i++) {
-        var message = JSON.parse(chunks[i]);
-
+      for (const message of parseChannelMessages(channel, pool)) {
         // There will be at most one NODE_HANDLE message in every chunk we
         // read because SCM_RIGHTS messages don't get coalesced. Make sure
         // that we deliver the handle with the right message however.
@@ -531,9 +579,6 @@ function setupChannel(target, channel) {
           handleMessage(message, undefined, false);
         }
       }
-      jsonBuffer = incompleteChunk;
-      this.buffering = jsonBuffer.length !== 0;
-
     } else {
       this.buffering = false;
       target.disconnect();
@@ -544,10 +589,10 @@ function setupChannel(target, channel) {
     }
   };
 
-  // object where socket lists will live
+  // Object where socket lists will live
   channel.sockets = { got: {}, send: {} };
 
-  // handlers will go through this
+  // Handlers will go through this
   target.on('internalMessage', function(message, handle) {
     // Once acknowledged - continue sending handles.
     if (message.cmd === 'NODE_HANDLE_ACK' ||
@@ -564,8 +609,8 @@ function setupChannel(target, channel) {
         }
       }
 
-      assert(Array.isArray(target._handleQueue));
-      var queue = target._handleQueue;
+      assert(ArrayIsArray(target._handleQueue));
+      const queue = target._handleQueue;
       target._handleQueue = null;
 
       if (target._pendingMessage) {
@@ -575,8 +620,8 @@ function setupChannel(target, channel) {
                      target._pendingMessage.callback);
       }
 
-      for (var i = 0; i < queue.length; i++) {
-        var args = queue[i];
+      for (let i = 0; i < queue.length; i++) {
+        const args = queue[i];
         target._send(args.message, args.handle, args.options, args.callback);
       }
 
@@ -601,16 +646,15 @@ function setupChannel(target, channel) {
     // a message.
     target._send({ cmd: 'NODE_HANDLE_ACK' }, null, true);
 
-    var obj = handleConversion[message.type];
+    const obj = handleConversion[message.type];
 
     // Update simultaneous accepts on Windows
     if (process.platform === 'win32') {
-      handle._simultaneousAccepts = false;
-      net._setSimultaneousAccepts(handle);
+      handle.setSimultaneousAccepts(false);
     }
 
     // Convert handle object
-    obj.got.call(this, message, handle, function(handle) {
+    obj.got.call(this, message, handle, (handle) => {
       handleMessage(message.msg, handle, isInternal(message.msg));
     });
   });
@@ -628,7 +672,7 @@ function setupChannel(target, channel) {
       throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
     }
 
-    options = Object.assign({ swallowErrors: false }, options);
+    options = { swallowErrors: false, ...options };
 
     if (this.connected) {
       return this._send(message, handle, options, callback);
@@ -648,14 +692,28 @@ function setupChannel(target, channel) {
     if (message === undefined)
       throw new ERR_MISSING_ARGS('message');
 
+    // Non-serializable messages should not reach the remote
+    // end point; as any failure in the stringification there
+    // will result in error message that is weakly consumable.
+    // So perform a sanity check on message prior to sending.
+    if (typeof message !== 'string' &&
+        typeof message !== 'object' &&
+        typeof message !== 'number' &&
+        typeof message !== 'boolean') {
+      throw new ERR_INVALID_ARG_TYPE(
+        'message', ['string', 'object', 'number', 'boolean'], message);
+    }
+
     // Support legacy function signature
     if (typeof options === 'boolean') {
       options = { swallowErrors: options };
     }
 
-    // package messages with a handle object
+    let obj;
+
+    // Package messages with a handle object
     if (handle) {
-      // this message will be handled by an internalMessage event handler
+      // This message will be handled by an internalMessage event handler
       message = {
         cmd: 'NODE_HANDLE',
         type: null,
@@ -687,7 +745,7 @@ function setupChannel(target, channel) {
         return this._handleQueue.length === 1;
       }
 
-      var obj = handleConversion[message.type];
+      obj = handleConversion[message.type];
 
       // convert TCP object to native handle object
       handle = handleConversion[message.type].send.call(target,
@@ -701,8 +759,8 @@ function setupChannel(target, channel) {
         message = message.msg;
 
       // Update simultaneous accepts on Windows
-      if (obj.simultaneousAccepts) {
-        net._setSimultaneousAccepts(handle);
+      if (obj.simultaneousAccepts && process.platform === 'win32') {
+        handle.setSimultaneousAccepts(true);
       }
     } else if (this._handleQueue &&
                !(message && (message.cmd === 'NODE_HANDLE_ACK' ||
@@ -717,11 +775,10 @@ function setupChannel(target, channel) {
       return this._handleQueue.length === 1;
     }
 
-    var req = new WriteWrap();
-    req.async = false;
+    const req = new WriteWrap();
 
-    var string = JSON.stringify(message) + '\n';
-    var err = channel.writeUtf8String(req, string, handle);
+    const err = writeChannelMessage(channel, req, message, handle);
+    const wasAsyncWrite = streamBaseState[kLastWriteWasAsync];
 
     if (err === 0) {
       if (handle) {
@@ -731,8 +788,8 @@ function setupChannel(target, channel) {
           obj.postSend(message, handle, options, callback, target);
       }
 
-      if (req.async) {
-        req.oncomplete = function() {
+      if (wasAsyncWrite) {
+        req.oncomplete = () => {
           control.unref();
           if (typeof callback === 'function')
             callback(null);
@@ -760,7 +817,7 @@ function setupChannel(target, channel) {
     return channel.writeQueueSize < (65536 * 2);
   };
 
-  // connected will be set to false immediately when a disconnect() is
+  // Connected will be set to false immediately when a disconnect() is
   // requested, even though the channel might still be alive internally to
   // process queued messages. The three states are distinguished as follows:
   // - disconnect() never requested: channel is not null and connected
@@ -796,7 +853,7 @@ function setupChannel(target, channel) {
     if (this._pendingMessage)
       closePendingHandle(this);
 
-    var fired = false;
+    let fired = false;
     function finish() {
       if (fired) return;
       fired = true;
@@ -824,7 +881,7 @@ function setupChannel(target, channel) {
     if (!target.channel)
       return;
 
-    var eventName = (internal ? 'internalMessage' : 'message');
+    const eventName = (internal ? 'internalMessage' : 'message');
 
     process.nextTick(emit, eventName, message, handle);
   }
@@ -844,34 +901,28 @@ function isInternal(message) {
 
 function nop() { }
 
-function _validateStdio(stdio, sync) {
-  var ipc;
-  var ipcFd;
+function getValidStdio(stdio, sync) {
+  let ipc;
+  let ipcFd;
 
   // Replace shortcut with an array
   if (typeof stdio === 'string') {
-    switch (stdio) {
-      case 'ignore': stdio = ['ignore', 'ignore', 'ignore']; break;
-      case 'pipe': stdio = ['pipe', 'pipe', 'pipe']; break;
-      case 'inherit': stdio = [0, 1, 2]; break;
-      default:
-        throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
-    }
-  } else if (!Array.isArray(stdio)) {
-    throw new ERR_INVALID_OPT_VALUE('stdio', util.inspect(stdio));
+    stdio = stdioStringToArray(stdio);
+  } else if (!ArrayIsArray(stdio)) {
+    throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
   }
 
   // At least 3 stdio will be created
   // Don't concat() a new Array() because it would be sparse, and
   // stdio.reduce() would skip the sparse elements of stdio.
-  // See http://stackoverflow.com/a/5501711/3561
+  // See https://stackoverflow.com/a/5501711/3561
   while (stdio.length < 3) stdio.push(undefined);
 
   // Translate stdio into C++-readable form
   // (i.e. PipeWraps or fds)
-  stdio = stdio.reduce(function(acc, stdio, i) {
+  stdio = stdio.reduce((acc, stdio, i) => {
     function cleanup() {
-      for (var i = 0; i < acc.length; i++) {
+      for (let i = 0; i < acc.length; i++) {
         if ((acc[i].type === 'pipe' || acc[i].type === 'ipc') && acc[i].handle)
           acc[i].handle.close();
       }
@@ -884,8 +935,8 @@ function _validateStdio(stdio, sync) {
 
     if (stdio === 'ignore') {
       acc.push({ type: 'ignore' });
-    } else if (stdio === 'pipe' || typeof stdio === 'number' && stdio < 0) {
-      var a = {
+    } else if (stdio === 'pipe' || (typeof stdio === 'number' && stdio < 0)) {
+      const a = {
         type: 'pipe',
         readable: i === 0,
         writable: i !== 0
@@ -925,24 +976,25 @@ function _validateStdio(stdio, sync) {
       });
     } else if (getHandleWrapType(stdio) || getHandleWrapType(stdio.handle) ||
                getHandleWrapType(stdio._handle)) {
-      var handle = getHandleWrapType(stdio) ?
+      const handle = getHandleWrapType(stdio) ?
         stdio :
         getHandleWrapType(stdio.handle) ? stdio.handle : stdio._handle;
 
       acc.push({
         type: 'wrap',
         wrapType: getHandleWrapType(handle),
-        handle: handle
+        handle: handle,
+        _stdio: stdio
       });
     } else if (isArrayBufferView(stdio) || typeof stdio === 'string') {
       if (!sync) {
         cleanup();
-        throw new ERR_INVALID_SYNC_FORK_INPUT(util.inspect(stdio));
+        throw new ERR_INVALID_SYNC_FORK_INPUT(inspect(stdio));
       }
     } else {
       // Cleanup
       cleanup();
-      throw new ERR_INVALID_OPT_VALUE('stdio', util.inspect(stdio));
+      throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
     }
 
     return acc;
@@ -953,10 +1005,10 @@ function _validateStdio(stdio, sync) {
 
 
 function getSocketList(type, worker, key) {
-  var sockets = worker.channel.sockets[type];
-  var socketList = sockets[key];
+  const sockets = worker.channel.sockets[type];
+  let socketList = sockets[key];
   if (!socketList) {
-    var Construct = type === 'send' ? SocketListSend : SocketListReceive;
+    const Construct = type === 'send' ? SocketListSend : SocketListReceive;
     socketList = sockets[key] = new Construct(worker, key);
   }
   return socketList;
@@ -972,11 +1024,11 @@ function maybeClose(subprocess) {
 }
 
 function spawnSync(opts) {
-  var options = opts.options;
-  var result = spawn_sync.spawn(options);
+  const options = opts.options;
+  const result = spawn_sync.spawn(options);
 
   if (result.output && options.encoding && options.encoding !== 'buffer') {
-    for (var i = 0; i < result.output.length; i++) {
+    for (let i = 0; i < result.output.length; i++) {
       if (!result.output[i])
         continue;
       result.output[i] = result.output[i].toString(options.encoding);
@@ -998,6 +1050,7 @@ function spawnSync(opts) {
 module.exports = {
   ChildProcess,
   setupChannel,
-  _validateStdio,
+  getValidStdio,
+  stdioStringToArray,
   spawnSync
 };
