@@ -1,52 +1,70 @@
 'use strict';
 
+const {
+  MathMax,
+  MathMin,
+  NumberIsSafeInteger,
+  Symbol,
+} = primordials;
+
 const {
   F_OK,
   O_SYMLINK,
   O_WRONLY,
   S_IFMT,
   S_IFREG
-} = process.binding('constants').fs;
-const binding = process.binding('fs');
+} = internalBinding('constants').fs;
+const binding = internalBinding('fs');
 const { Buffer, kMaxLength } = require('buffer');
 const {
   ERR_FS_FILE_TOO_LARGE,
   ERR_INVALID_ARG_TYPE,
+  ERR_INVALID_ARG_VALUE,
   ERR_METHOD_NOT_IMPLEMENTED
 } = require('internal/errors').codes;
-const { toPathIfFileURL } = require('internal/url');
 const { isUint8Array } = require('internal/util/types');
+const { rimrafPromises } = require('internal/fs/rimraf');
 const {
   copyObject,
   getDirents,
   getOptions,
   getStatsFromBinding,
+  getValidatedPath,
   nullCheck,
   preprocessSymlinkDestination,
   stringToFlags,
   stringToSymlinkType,
   toUnixTimestamp,
-  validateBuffer,
+  validateBufferArray,
   validateOffsetLengthRead,
   validateOffsetLengthWrite,
-  validatePath
+  validateRmdirOptions,
+  warnOnNonPortableTemplate
 } = require('internal/fs/utils');
+const { opendir } = require('internal/fs/dir');
 const {
+  parseMode,
+  validateBuffer,
   validateInteger,
-  validateMode,
   validateUint32
 } = require('internal/validators');
 const pathModule = require('path');
 const { promisify } = require('internal/util');
 
-const kHandle = Symbol('handle');
+const kHandle = Symbol('kHandle');
+const kFd = Symbol('kFd');
 const { kUsePromises } = binding;
+const {
+  JSTransferable, kDeserialize, kTransfer, kTransferList
+} = require('internal/worker/js_transferable');
 
 const getDirectoryEntriesPromise = promisify(getDirents);
 
-class FileHandle {
+class FileHandle extends JSTransferable {
   constructor(filehandle) {
+    super();
     this[kHandle] = filehandle;
+    this[kFd] = filehandle ? filehandle.fd : -1;
   }
 
   getAsyncId() {
@@ -54,11 +72,11 @@ class FileHandle {
   }
 
   get fd() {
-    return this[kHandle].fd;
+    return this[kFd];
   }
 
   appendFile(data, options) {
-    return appendFile(this, data, options);
+    return writeFile(this, data, options);
   }
 
   chmod(mode) {
@@ -81,6 +99,10 @@ class FileHandle {
     return read(this, buffer, offset, length, position);
   }
 
+  readv(buffers, position) {
+    return readv(this, buffers, position);
+  }
+
   readFile(options) {
     return readFile(this, options);
   }
@@ -101,13 +123,38 @@ class FileHandle {
     return write(this, buffer, offset, length, position);
   }
 
+  writev(buffers, position) {
+    return writev(this, buffers, position);
+  }
+
   writeFile(data, options) {
     return writeFile(this, data, options);
   }
 
-  close() {
+  close = () => {
+    this[kFd] = -1;
     return this[kHandle].close();
   }
+
+  [kTransfer]() {
+    const handle = this[kHandle];
+    this[kFd] = -1;
+    this[kHandle] = null;
+
+    return {
+      data: { handle },
+      deserializeInfo: 'internal/fs/promises:FileHandle'
+    };
+  }
+
+  [kTransferList]() {
+    return [ this[kHandle] ];
+  }
+
+  [kDeserialize]({ handle }) {
+    this[kHandle] = handle;
+    this[kFd] = handle.fd;
+  }
 }
 
 function validateFileHandle(handle) {
@@ -123,7 +170,7 @@ async function writeFileHandle(filehandle, data, options) {
   do {
     const { bytesWritten } =
       await write(filehandle, buffer, 0,
-                  Math.min(16384, buffer.length));
+                  MathMin(16384, buffer.length));
     remaining -= bytesWritten;
     buffer = buffer.slice(bytesWritten);
   } while (remaining > 0);
@@ -149,7 +196,7 @@ async function readFileHandle(filehandle, options) {
   const chunks = [];
   const chunkSize = size === 0 ?
     kReadFileMaxChunkSize :
-    Math.min(size, kReadFileMaxChunkSize);
+    MathMin(size, kReadFileMaxChunkSize);
   let endOfFile = false;
   do {
     const buf = Buffer.alloc(chunkSize);
@@ -161,18 +208,14 @@ async function readFileHandle(filehandle, options) {
   } while (!endOfFile);
 
   const result = Buffer.concat(chunks);
-  if (options.encoding) {
-    return result.toString(options.encoding);
-  } else {
-    return result;
-  }
+
+  return options.encoding ? result.toString(options.encoding) : result;
 }
 
 // All of the functions are defined as async in order to ensure that errors
 // thrown cause promise rejections rather than being thrown synchronously.
 async function access(path, mode = F_OK) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
 
   mode = mode | 0;
   return binding.access(pathModule.toNamespacedPath(path), mode,
@@ -180,10 +223,8 @@ async function access(path, mode = F_OK) {
 }
 
 async function copyFile(src, dest, flags) {
-  src = toPathIfFileURL(src);
-  dest = toPathIfFileURL(dest);
-  validatePath(src, 'src');
-  validatePath(dest, 'dest');
+  src = getValidatedPath(src, 'src');
+  dest = getValidatedPath(dest, 'dest');
   flags = flags | 0;
   return binding.copyFile(pathModule.toNamespacedPath(src),
                           pathModule.toNamespacedPath(dest),
@@ -193,11 +234,10 @@ async function copyFile(src, dest, flags) {
 // Note that unlike fs.open() which uses numeric file descriptors,
 // fsPromises.open() uses the fs.FileHandle class.
 async function open(path, flags, mode) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   if (arguments.length < 2) flags = 'r';
   const flagsNumber = stringToFlags(flags);
-  mode = validateMode(mode, 'mode', 0o666);
+  mode = parseMode(mode, 'mode', 0o666);
   return new FileHandle(
     await binding.openFileHandle(pathModule.toNamespacedPath(path),
                                  flagsNumber, mode, kUsePromises));
@@ -213,9 +253,14 @@ async function read(handle, buffer, offset, length, position) {
   if (length === 0)
     return { bytesRead: length, buffer };
 
+  if (buffer.length === 0) {
+    throw new ERR_INVALID_ARG_VALUE('buffer', buffer,
+                                    'is empty and cannot be written');
+  }
+
   validateOffsetLengthRead(offset, length, buffer.length);
 
-  if (!Number.isSafeInteger(position))
+  if (!NumberIsSafeInteger(position))
     position = -1;
 
   const bytesRead = (await binding.read(handle.fd, buffer, offset, length,
@@ -224,6 +269,18 @@ async function read(handle, buffer, offset, length, position) {
   return { bytesRead, buffer };
 }
 
+async function readv(handle, buffers, position) {
+  validateFileHandle(handle);
+  validateBufferArray(buffers);
+
+  if (typeof position !== 'number')
+    position = null;
+
+  const bytesRead = (await binding.readBuffers(handle.fd, buffers, position,
+                                               kUsePromises)) || 0;
+  return { bytesRead, buffers };
+}
+
 async function write(handle, buffer, offset, length, position) {
   validateFileHandle(handle);
 
@@ -251,31 +308,47 @@ async function write(handle, buffer, offset, length, position) {
   return { bytesWritten, buffer };
 }
 
+async function writev(handle, buffers, position) {
+  validateFileHandle(handle);
+  validateBufferArray(buffers);
+
+  if (typeof position !== 'number')
+    position = null;
+
+  const bytesWritten = (await binding.writeBuffers(handle.fd, buffers, position,
+                                                   kUsePromises)) || 0;
+  return { bytesWritten, buffers };
+}
+
 async function rename(oldPath, newPath) {
-  oldPath = toPathIfFileURL(oldPath);
-  newPath = toPathIfFileURL(newPath);
-  validatePath(oldPath, 'oldPath');
-  validatePath(newPath, 'newPath');
+  oldPath = getValidatedPath(oldPath, 'oldPath');
+  newPath = getValidatedPath(newPath, 'newPath');
   return binding.rename(pathModule.toNamespacedPath(oldPath),
                         pathModule.toNamespacedPath(newPath),
                         kUsePromises);
 }
 
 async function truncate(path, len = 0) {
-  return ftruncate(await open(path, 'r+'), len);
+  const fd = await open(path, 'r+');
+  return ftruncate(fd, len).finally(fd.close.bind(fd));
 }
 
 async function ftruncate(handle, len = 0) {
   validateFileHandle(handle);
   validateInteger(len, 'len');
-  len = Math.max(0, len);
+  len = MathMax(0, len);
   return binding.ftruncate(handle.fd, len, kUsePromises);
 }
 
-async function rmdir(path) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
-  return binding.rmdir(pathModule.toNamespacedPath(path), kUsePromises);
+async function rmdir(path, options) {
+  path = pathModule.toNamespacedPath(getValidatedPath(path));
+  options = validateRmdirOptions(options);
+
+  if (options.recursive) {
+    return rimrafPromises(path, options);
+  }
+
+  return binding.rmdir(path, kUsePromises);
 }
 
 async function fdatasync(handle) {
@@ -296,21 +369,18 @@ async function mkdir(path, options) {
     recursive = false,
     mode = 0o777
   } = options || {};
-  path = toPathIfFileURL(path);
-
-  validatePath(path);
+  path = getValidatedPath(path);
   if (typeof recursive !== 'boolean')
-    throw new ERR_INVALID_ARG_TYPE('recursive', 'boolean', recursive);
+    throw new ERR_INVALID_ARG_TYPE('options.recursive', 'boolean', recursive);
 
   return binding.mkdir(pathModule.toNamespacedPath(path),
-                       validateMode(mode, 'mode', 0o777), recursive,
+                       parseMode(mode, 'mode', 0o777), recursive,
                        kUsePromises);
 }
 
 async function readdir(path, options) {
   options = getOptions(options, {});
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   const result = await binding.readdir(pathModule.toNamespacedPath(path),
                                        options.encoding,
                                        !!options.withFileTypes,
@@ -322,18 +392,15 @@ async function readdir(path, options) {
 
 async function readlink(path, options) {
   options = getOptions(options, {});
-  path = toPathIfFileURL(path);
-  validatePath(path, 'oldPath');
+  path = getValidatedPath(path, 'oldPath');
   return binding.readlink(pathModule.toNamespacedPath(path),
                           options.encoding, kUsePromises);
 }
 
 async function symlink(target, path, type_) {
   const type = (typeof type_ === 'string' ? type_ : null);
-  target = toPathIfFileURL(target);
-  path = toPathIfFileURL(path);
-  validatePath(target, 'target');
-  validatePath(path);
+  target = getValidatedPath(target, 'target');
+  path = getValidatedPath(path);
   return binding.symlink(preprocessSymlinkDestination(target, type, path),
                          pathModule.toNamespacedPath(path),
                          stringToSymlinkType(type),
@@ -347,47 +414,41 @@ async function fstat(handle, options = { bigint: false }) {
 }
 
 async function lstat(path, options = { bigint: false }) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   const result = await binding.lstat(pathModule.toNamespacedPath(path),
                                      options.bigint, kUsePromises);
   return getStatsFromBinding(result);
 }
 
 async function stat(path, options = { bigint: false }) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   const result = await binding.stat(pathModule.toNamespacedPath(path),
                                     options.bigint, kUsePromises);
   return getStatsFromBinding(result);
 }
 
 async function link(existingPath, newPath) {
-  existingPath = toPathIfFileURL(existingPath);
-  newPath = toPathIfFileURL(newPath);
-  validatePath(existingPath, 'existingPath');
-  validatePath(newPath, 'newPath');
+  existingPath = getValidatedPath(existingPath, 'existingPath');
+  newPath = getValidatedPath(newPath, 'newPath');
   return binding.link(pathModule.toNamespacedPath(existingPath),
                       pathModule.toNamespacedPath(newPath),
                       kUsePromises);
 }
 
 async function unlink(path) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   return binding.unlink(pathModule.toNamespacedPath(path), kUsePromises);
 }
 
 async function fchmod(handle, mode) {
   validateFileHandle(handle);
-  mode = validateMode(mode, 'mode');
+  mode = parseMode(mode, 'mode');
   return binding.fchmod(handle.fd, mode, kUsePromises);
 }
 
 async function chmod(path, mode) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
-  mode = validateMode(mode, 'mode');
+  path = getValidatedPath(path);
+  mode = parseMode(mode, 'mode');
   return binding.chmod(pathModule.toNamespacedPath(path), mode, kUsePromises);
 }
 
@@ -396,12 +457,11 @@ async function lchmod(path, mode) {
     throw new ERR_METHOD_NOT_IMPLEMENTED('lchmod()');
 
   const fd = await open(path, O_WRONLY | O_SYMLINK);
-  return fchmod(fd, mode).finally(fd.close.bind(fd));
+  return fchmod(fd, mode).finally(fd.close);
 }
 
 async function lchown(path, uid, gid) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   validateUint32(uid, 'uid');
   validateUint32(gid, 'gid');
   return binding.lchown(pathModule.toNamespacedPath(path),
@@ -416,8 +476,7 @@ async function fchown(handle, uid, gid) {
 }
 
 async function chown(path, uid, gid) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   validateUint32(uid, 'uid');
   validateUint32(gid, 'gid');
   return binding.chown(pathModule.toNamespacedPath(path),
@@ -425,8 +484,7 @@ async function chown(path, uid, gid) {
 }
 
 async function utimes(path, atime, mtime) {
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   return binding.utimes(pathModule.toNamespacedPath(path),
                         toUnixTimestamp(atime),
                         toUnixTimestamp(mtime),
@@ -440,10 +498,17 @@ async function futimes(handle, atime, mtime) {
   return binding.futimes(handle.fd, atime, mtime, kUsePromises);
 }
 
+async function lutimes(path, atime, mtime) {
+  path = getValidatedPath(path);
+  return binding.lutimes(pathModule.toNamespacedPath(path),
+                         toUnixTimestamp(atime),
+                         toUnixTimestamp(mtime),
+                         kUsePromises);
+}
+
 async function realpath(path, options) {
   options = getOptions(options, {});
-  path = toPathIfFileURL(path);
-  validatePath(path);
+  path = getValidatedPath(path);
   return binding.realpath(path, options.encoding, kUsePromises);
 }
 
@@ -453,6 +518,7 @@ async function mkdtemp(prefix, options) {
     throw new ERR_INVALID_ARG_TYPE('prefix', 'string', prefix);
   }
   nullCheck(prefix);
+  warnOnNonPortableTemplate(prefix);
   return binding.mkdtemp(`${prefix}XXXXXX`, options.encoding, kUsePromises);
 }
 
@@ -464,7 +530,7 @@ async function writeFile(path, data, options) {
     return writeFileHandle(path, data, options);
 
   const fd = await open(path, flag, options.mode);
-  return writeFileHandle(fd, data, options).finally(fd.close.bind(fd));
+  return writeFileHandle(fd, data, options).finally(fd.close);
 }
 
 async function appendFile(path, data, options) {
@@ -482,32 +548,38 @@ async function readFile(path, options) {
     return readFileHandle(path, options);
 
   const fd = await open(path, flag, 0o666);
-  return readFileHandle(fd, options).finally(fd.close.bind(fd));
+  return readFileHandle(fd, options).finally(fd.close);
 }
 
 module.exports = {
-  access,
-  copyFile,
-  open,
-  rename,
-  truncate,
-  rmdir,
-  mkdir,
-  readdir,
-  readlink,
-  symlink,
-  lstat,
-  stat,
-  link,
-  unlink,
-  chmod,
-  lchmod,
-  lchown,
-  chown,
-  utimes,
-  realpath,
-  mkdtemp,
-  writeFile,
-  appendFile,
-  readFile
+  exports: {
+    access,
+    copyFile,
+    open,
+    opendir: promisify(opendir),
+    rename,
+    truncate,
+    rmdir,
+    mkdir,
+    readdir,
+    readlink,
+    symlink,
+    lstat,
+    stat,
+    link,
+    unlink,
+    chmod,
+    lchmod,
+    lchown,
+    chown,
+    utimes,
+    lutimes,
+    realpath,
+    mkdtemp,
+    writeFile,
+    appendFile,
+    readFile,
+  },
+
+  FileHandle
 };
