@@ -1,15 +1,32 @@
 'use strict';
 
+const {
+  ArrayIsArray,
+  Error,
+  MathMax,
+  Number,
+  ObjectCreate,
+  ObjectKeys,
+  Set,
+  Symbol,
+} = primordials;
+
 const binding = internalBinding('http2');
 const {
-  ERR_HTTP2_HEADER_SINGLE_VALUE,
-  ERR_HTTP2_INVALID_CONNECTION_HEADERS,
-  ERR_HTTP2_INVALID_PSEUDOHEADER,
-  ERR_HTTP2_INVALID_SETTING_VALUE,
-  ERR_INVALID_ARG_TYPE
-} = require('internal/errors').codes;
+  codes: {
+    ERR_HTTP2_HEADER_SINGLE_VALUE,
+    ERR_HTTP2_INVALID_CONNECTION_HEADERS,
+    ERR_HTTP2_INVALID_PSEUDOHEADER,
+    ERR_HTTP2_INVALID_SETTING_VALUE,
+    ERR_INVALID_ARG_TYPE
+  },
+  addCodeToName,
+  hideStackFrames
+} = require('internal/errors');
 
 const kSocket = Symbol('socket');
+const kProxySocket = Symbol('proxySocket');
+const kRequest = Symbol('request');
 
 const {
   NGHTTP2_SESSION_CLIENT,
@@ -186,10 +203,11 @@ const IDX_OPTIONS_MAX_HEADER_LIST_PAIRS = 5;
 const IDX_OPTIONS_MAX_OUTSTANDING_PINGS = 6;
 const IDX_OPTIONS_MAX_OUTSTANDING_SETTINGS = 7;
 const IDX_OPTIONS_MAX_SESSION_MEMORY = 8;
-const IDX_OPTIONS_FLAGS = 9;
+const IDX_OPTIONS_MAX_SETTINGS = 9;
+const IDX_OPTIONS_FLAGS = 10;
 
 function updateOptionsBuffer(options) {
-  var flags = 0;
+  let flags = 0;
   if (typeof options.maxDeflateDynamicTableSize === 'number') {
     flags |= (1 << IDX_OPTIONS_MAX_DEFLATE_DYNAMIC_TABLE_SIZE);
     optionsBuffer[IDX_OPTIONS_MAX_DEFLATE_DYNAMIC_TABLE_SIZE] =
@@ -228,12 +246,17 @@ function updateOptionsBuffer(options) {
   if (typeof options.maxOutstandingSettings === 'number') {
     flags |= (1 << IDX_OPTIONS_MAX_OUTSTANDING_SETTINGS);
     optionsBuffer[IDX_OPTIONS_MAX_OUTSTANDING_SETTINGS] =
-      Math.max(1, options.maxOutstandingSettings);
+      MathMax(1, options.maxOutstandingSettings);
   }
   if (typeof options.maxSessionMemory === 'number') {
     flags |= (1 << IDX_OPTIONS_MAX_SESSION_MEMORY);
     optionsBuffer[IDX_OPTIONS_MAX_SESSION_MEMORY] =
-      Math.max(1, options.maxSessionMemory);
+      MathMax(1, options.maxSessionMemory);
+  }
+  if (typeof options.maxSettings === 'number') {
+    flags |= (1 << IDX_OPTIONS_MAX_SETTINGS);
+    optionsBuffer[IDX_OPTIONS_MAX_SETTINGS] =
+      MathMax(1, options.maxSettings);
   }
   optionsBuffer[IDX_OPTIONS_FLAGS] = flags;
 }
@@ -241,7 +264,7 @@ function updateOptionsBuffer(options) {
 function getDefaultSettings() {
   settingsBuffer[IDX_SETTINGS_FLAGS] = 0;
   binding.refreshDefaultSettings();
-  const holder = Object.create(null);
+  const holder = ObjectCreate(null);
 
   const flags = settingsBuffer[IDX_SETTINGS_FLAGS];
 
@@ -277,20 +300,20 @@ function getDefaultSettings() {
 
   if ((flags & (1 << IDX_SETTINGS_MAX_HEADER_LIST_SIZE)) ===
       (1 << IDX_SETTINGS_MAX_HEADER_LIST_SIZE)) {
-    holder.maxHeaderListSize =
+    holder.maxHeaderListSize = holder.maxHeaderSize =
       settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE];
   }
 
   if ((flags & (1 << IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL)) ===
       (1 << IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL)) {
     holder.enableConnectProtocol =
-      settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL];
+      settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL] === 1;
   }
 
   return holder;
 }
 
-// remote is a boolean. true to fetch remote settings, false to fetch local.
+// Remote is a boolean. true to fetch remote settings, false to fetch local.
 // this is only called internally
 function getSettings(session, remote) {
   if (remote)
@@ -305,12 +328,14 @@ function getSettings(session, remote) {
     maxFrameSize: settingsBuffer[IDX_SETTINGS_MAX_FRAME_SIZE],
     maxConcurrentStreams: settingsBuffer[IDX_SETTINGS_MAX_CONCURRENT_STREAMS],
     maxHeaderListSize: settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE],
-    enableConnectProtocol: settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL]
+    maxHeaderSize: settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE],
+    enableConnectProtocol:
+      !!settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL]
   };
 }
 
 function updateSettingsBuffer(settings) {
-  var flags = 0;
+  let flags = 0;
   if (typeof settings.headerTableSize === 'number') {
     flags |= (1 << IDX_SETTINGS_HEADER_TABLE_SIZE);
     settingsBuffer[IDX_SETTINGS_HEADER_TABLE_SIZE] =
@@ -331,10 +356,20 @@ function updateSettingsBuffer(settings) {
     settingsBuffer[IDX_SETTINGS_MAX_FRAME_SIZE] =
       settings.maxFrameSize;
   }
-  if (typeof settings.maxHeaderListSize === 'number') {
+  if (typeof settings.maxHeaderListSize === 'number' ||
+      typeof settings.maxHeaderSize === 'number') {
     flags |= (1 << IDX_SETTINGS_MAX_HEADER_LIST_SIZE);
-    settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE] =
-      settings.maxHeaderListSize;
+    if (settings.maxHeaderSize !== undefined &&
+      (settings.maxHeaderSize !== settings.maxHeaderListSize)) {
+      process.emitWarning(
+        'settings.maxHeaderSize overwrite settings.maxHeaderListSize'
+      );
+      settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE] =
+        settings.maxHeaderSize;
+    } else {
+      settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE] =
+        settings.maxHeaderListSize;
+    }
   }
   if (typeof settings.enablePush === 'boolean') {
     flags |= (1 << IDX_SETTINGS_ENABLE_PUSH);
@@ -402,48 +437,42 @@ function isIllegalConnectionSpecificHeader(name, value) {
   }
 }
 
-function assertValidPseudoHeader(key) {
+const assertValidPseudoHeader = hideStackFrames((key) => {
   if (!kValidPseudoHeaders.has(key)) {
-    const err = new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
-    Error.captureStackTrace(err, assertValidPseudoHeader);
-    return err;
+    throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
   }
-}
+});
 
-function assertValidPseudoHeaderResponse(key) {
+const assertValidPseudoHeaderResponse = hideStackFrames((key) => {
   if (key !== ':status') {
-    const err = new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
-    Error.captureStackTrace(err, assertValidPseudoHeaderResponse);
-    return err;
+    throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
   }
-}
+});
 
-function assertValidPseudoHeaderTrailer(key) {
-  const err = new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
-  Error.captureStackTrace(err, assertValidPseudoHeaderTrailer);
-  return err;
-}
+const assertValidPseudoHeaderTrailer = hideStackFrames((key) => {
+  throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
+});
 
 function mapToHeaders(map,
                       assertValuePseudoHeader = assertValidPseudoHeader) {
   let ret = '';
   let count = 0;
-  const keys = Object.keys(map);
+  const keys = ObjectKeys(map);
   const singles = new Set();
-  let i;
+  let i, j;
   let isArray;
   let key;
   let value;
   let isSingleValueHeader;
   let err;
-  for (i = 0; i < keys.length; i++) {
+  for (i = 0; i < keys.length; ++i) {
     key = keys[i];
     value = map[key];
     if (value === undefined || key === '')
       continue;
     key = key.toLowerCase();
     isSingleValueHeader = kSingleValueHeaders.has(key);
-    isArray = Array.isArray(value);
+    isArray = ArrayIsArray(value);
     if (isArray) {
       switch (value.length) {
         case 0:
@@ -454,30 +483,30 @@ function mapToHeaders(map,
           break;
         default:
           if (isSingleValueHeader)
-            return new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
+            throw new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
       }
     } else {
       value = String(value);
     }
     if (isSingleValueHeader) {
       if (singles.has(key))
-        return new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
+        throw new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
       singles.add(key);
     }
     if (key[0] === ':') {
       err = assertValuePseudoHeader(key);
       if (err !== undefined)
-        return err;
+        throw err;
       ret = `${key}\0${value}\0${ret}`;
       count++;
       continue;
     }
     if (isIllegalConnectionSpecificHeader(key, value)) {
-      return new ERR_HTTP2_INVALID_CONNECTION_HEADERS(key);
+      throw new ERR_HTTP2_INVALID_CONNECTION_HEADERS(key);
     }
     if (isArray) {
-      for (var k = 0; k < value.length; k++) {
-        const val = String(value[k]);
+      for (j = 0; j < value.length; ++j) {
+        const val = String(value[j]);
         ret += `${key}\0${val}\0`;
       }
       count += value.length;
@@ -494,42 +523,42 @@ class NghttpError extends Error {
   constructor(ret) {
     super(binding.nghttp2ErrorString(ret));
     this.code = 'ERR_HTTP2_ERROR';
-    this.name = 'Error [ERR_HTTP2_ERROR]';
     this.errno = ret;
+    addCodeToName(this, super.name, 'ERR_HTTP2_ERROR');
+  }
+
+  toString() {
+    return `${this.name} [${this.code}]: ${this.message}`;
   }
 }
 
-function assertIsObject(value, name, types = 'Object') {
+const assertIsObject = hideStackFrames((value, name, types) => {
   if (value !== undefined &&
       (value === null ||
        typeof value !== 'object' ||
-       Array.isArray(value))) {
-    const err = new ERR_INVALID_ARG_TYPE(name, types, value);
-    Error.captureStackTrace(err, assertIsObject);
-    throw err;
+       ArrayIsArray(value))) {
+    throw new ERR_INVALID_ARG_TYPE(name, types || 'Object', value);
   }
-}
+});
 
-function assertWithinRange(name, value, min = 0, max = Infinity) {
-  if (value !== undefined &&
+const assertWithinRange = hideStackFrames(
+  (name, value, min = 0, max = Infinity) => {
+    if (value !== undefined &&
       (typeof value !== 'number' || value < min || value > max)) {
-    const err = new ERR_HTTP2_INVALID_SETTING_VALUE.RangeError(name, value);
-    err.min = min;
-    err.max = max;
-    err.actual = value;
-    Error.captureStackTrace(err, assertWithinRange);
-    throw err;
+      throw new ERR_HTTP2_INVALID_SETTING_VALUE.RangeError(
+        name, value, min, max);
+    }
   }
-}
+);
 
 function toHeaderObject(headers) {
-  const obj = Object.create(null);
-  for (var n = 0; n < headers.length; n = n + 2) {
-    var name = headers[n];
-    var value = headers[n + 1];
+  const obj = ObjectCreate(null);
+  for (var n = 0; n < headers.length; n += 2) {
+    const name = headers[n];
+    let value = headers[n + 1];
     if (name === HTTP2_HEADER_STATUS)
       value |= 0;
-    var existing = obj[name];
+    const existing = obj[name];
     if (existing === undefined) {
       obj[name] = name === HTTP2_HEADER_SET_COOKIE ? [value] : value;
     } else if (!kSingleValueHeaders.has(name)) {
@@ -592,6 +621,8 @@ module.exports = {
   getStreamState,
   isPayloadMeaningless,
   kSocket,
+  kProxySocket,
+  kRequest,
   mapToHeaders,
   NghttpError,
   sessionName,
