@@ -1,6 +1,18 @@
 'use strict';
 
 const {
+  ArrayIsArray,
+  Boolean,
+  NumberIsSafeInteger,
+  ObjectDefineProperties,
+  ObjectDefineProperty,
+  ObjectKeys,
+  Set,
+  Symbol,
+} = primordials;
+
+const {
+  ELDHistogram: _ELDHistogram,
   PerformanceEntry,
   mark: _mark,
   clearMark: _clearMark,
@@ -11,8 +23,11 @@ const {
   timeOrigin,
   timeOriginTimestamp,
   timerify,
-  constants
-} = process.binding('performance');
+  constants,
+  installGarbageCollectionTracking,
+  removeGarbageCollectionTracking,
+  loopIdleTime,
+} = internalBinding('performance');
 
 const {
   NODE_PERFORMANCE_ENTRY_TYPE_NODE,
@@ -21,6 +36,7 @@ const {
   NODE_PERFORMANCE_ENTRY_TYPE_GC,
   NODE_PERFORMANCE_ENTRY_TYPE_FUNCTION,
   NODE_PERFORMANCE_ENTRY_TYPE_HTTP2,
+  NODE_PERFORMANCE_ENTRY_TYPE_HTTP,
 
   NODE_PERFORMANCE_MILESTONE_NODE_START,
   NODE_PERFORMANCE_MILESTONE_V8_START,
@@ -33,8 +49,21 @@ const {
 const { AsyncResource } = require('async_hooks');
 const L = require('internal/linkedlist');
 const kInspect = require('internal/util').customInspectSymbol;
-const { inherits } = require('util');
 
+const {
+  ERR_INVALID_CALLBACK,
+  ERR_INVALID_ARG_TYPE,
+  ERR_INVALID_OPT_VALUE,
+  ERR_VALID_PERFORMANCE_ENTRY_TYPE,
+  ERR_INVALID_PERFORMANCE_MARK
+} = require('internal/errors').codes;
+
+const {
+  Histogram,
+  kHandle,
+} = require('internal/histogram');
+
+const { setImmediate } = require('timers');
 const kCallback = Symbol('callback');
 const kTypes = Symbol('types');
 const kEntries = Symbol('entries');
@@ -55,7 +84,8 @@ const observerableTypes = [
   'measure',
   'gc',
   'function',
-  'http2'
+  'http2',
+  'http'
 ];
 
 const IDX_STREAM_STATS_ID = 0;
@@ -75,14 +105,16 @@ const IDX_SESSION_STATS_DATA_SENT = 6;
 const IDX_SESSION_STATS_DATA_RECEIVED = 7;
 const IDX_SESSION_STATS_MAX_CONCURRENT_STREAMS = 8;
 
+let http2;
 let sessionStats;
 let streamStats;
 
 function collectHttp2Stats(entry) {
+  if (http2 === undefined) http2 = internalBinding('http2');
   switch (entry.name) {
     case 'Http2Stream':
       if (streamStats === undefined)
-        streamStats = process.binding('http2').streamStats;
+        streamStats = http2.streamStats;
       entry.id =
         streamStats[IDX_STREAM_STATS_ID] >>> 0;
       entry.timeToFirstByte =
@@ -98,7 +130,7 @@ function collectHttp2Stats(entry) {
       break;
     case 'Http2Session':
       if (sessionStats === undefined)
-        sessionStats = process.binding('http2').sessionStats;
+        sessionStats = http2.sessionStats;
       entry.type =
         sessionStats[IDX_SESSION_STATS_TYPE] >>> 0 === 0 ? 'server' : 'client';
       entry.pingRTT =
@@ -121,14 +153,6 @@ function collectHttp2Stats(entry) {
   }
 }
 
-
-let errors;
-function lazyErrors() {
-  if (errors === undefined)
-    errors = require('internal/errors').codes;
-  return errors;
-}
-
 function now() {
   const hr = process.hrtime();
   return hr[0] * 1000 + hr[1] / 1e6;
@@ -141,7 +165,7 @@ function getMilestoneTimestamp(milestoneIdx) {
   return ns / 1e6 - timeOrigin;
 }
 
-class PerformanceNodeTiming {
+class PerformanceNodeTiming extends PerformanceEntry {
   get name() {
     return 'node';
   }
@@ -182,6 +206,10 @@ class PerformanceNodeTiming {
     return getMilestoneTimestamp(NODE_PERFORMANCE_MILESTONE_BOOTSTRAP_COMPLETE);
   }
 
+  get idleTime() {
+    return loopIdleTime();
+  }
+
   [kInspect]() {
     return {
       name: 'node',
@@ -193,29 +221,17 @@ class PerformanceNodeTiming {
       bootstrapComplete: this.bootstrapComplete,
       environment: this.environment,
       loopStart: this.loopStart,
-      loopExit: this.loopExit,
-      thirdPartyMainStart: this.thirdPartyMainStart,
-      thirdPartyMainEnd: this.thirdPartyMainEnd,
-      clusterSetupStart: this.clusterSetupStart,
-      clusterSetupEnd: this.clusterSetupEnd,
-      moduleLoadStart: this.moduleLoadStart,
-      moduleLoadEnd: this.moduleLoadEnd,
-      preloadModuleLoadStart: this.preloadModuleLoadStart,
-      preloadModuleLoadEnd: this.preloadModuleLoadEnd
+      loopExit: this.loopExit
     };
   }
 }
-// Use this instead of Extends because we want PerformanceEntry in the
-// prototype chain but we do not want to use the PerformanceEntry
-// constructor for this.
-inherits(PerformanceNodeTiming, PerformanceEntry);
 
 const nodeTiming = new PerformanceNodeTiming();
 
 // Maintains a list of entries as a linked list stored in insertion order.
 class PerformanceObserverEntryList {
   constructor() {
-    Object.defineProperties(this, {
+    ObjectDefineProperties(this, {
       [kEntries]: {
         writable: true,
         enumerable: false,
@@ -277,11 +293,10 @@ class PerformanceObserverEntryList {
 class PerformanceObserver extends AsyncResource {
   constructor(callback) {
     if (typeof callback !== 'function') {
-      const errors = lazyErrors();
-      throw new errors.ERR_INVALID_CALLBACK();
+      throw new ERR_INVALID_CALLBACK(callback);
     }
     super('PerformanceObserver');
-    Object.defineProperties(this, {
+    ObjectDefineProperties(this, {
       [kTypes]: {
         enumerable: false,
         writable: true,
@@ -311,42 +326,51 @@ class PerformanceObserver extends AsyncResource {
   }
 
   disconnect() {
+    const observerCountsGC = observerCounts[NODE_PERFORMANCE_ENTRY_TYPE_GC];
     const types = this[kTypes];
-    const keys = Object.keys(types);
-    for (var n = 0; n < keys.length; n++) {
-      const item = types[keys[n]];
+    for (const key of ObjectKeys(types)) {
+      const item = types[key];
       if (item) {
         L.remove(item);
-        observerCounts[keys[n]]--;
+        observerCounts[key]--;
       }
     }
     this[kTypes] = {};
+    if (observerCountsGC === 1 &&
+      observerCounts[NODE_PERFORMANCE_ENTRY_TYPE_GC] === 0) {
+      removeGarbageCollectionTracking();
+    }
   }
 
   observe(options) {
-    const errors = lazyErrors();
-    if (typeof options !== 'object' || options == null) {
-      throw new errors.ERR_INVALID_ARG_TYPE('options', 'Object', options);
+    if (typeof options !== 'object' || options === null) {
+      throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
     }
-    if (!Array.isArray(options.entryTypes)) {
-      throw new errors.ERR_INVALID_OPT_VALUE('entryTypes', options);
+    const { entryTypes } = options;
+    if (!ArrayIsArray(entryTypes)) {
+      throw new ERR_INVALID_OPT_VALUE('entryTypes', entryTypes);
     }
-    const entryTypes = options.entryTypes.filter(filterTypes).map(mapTypes);
-    if (entryTypes.length === 0) {
-      throw new errors.ERR_VALID_PERFORMANCE_ENTRY_TYPE();
+    const filteredEntryTypes = entryTypes.filter(filterTypes).map(mapTypes);
+    if (filteredEntryTypes.length === 0) {
+      throw new ERR_VALID_PERFORMANCE_ENTRY_TYPE();
     }
     this.disconnect();
+    const observerCountsGC = observerCounts[NODE_PERFORMANCE_ENTRY_TYPE_GC];
     this[kBuffer][kEntries] = [];
     L.init(this[kBuffer][kEntries]);
     this[kBuffering] = Boolean(options.buffered);
-    for (var n = 0; n < entryTypes.length; n++) {
-      const entryType = entryTypes[n];
+    for (const entryType of filteredEntryTypes) {
       const list = getObserversList(entryType);
+      if (this[kTypes][entryType]) continue;
       const item = { obs: this };
       this[kTypes][entryType] = item;
       L.append(list, item);
       observerCounts[entryType]++;
     }
+    if (observerCountsGC === 0 &&
+      observerCounts[NODE_PERFORMANCE_ENTRY_TYPE_GC] === 1) {
+      installGarbageCollectionTracking();
+    }
   }
 }
 
@@ -377,19 +401,20 @@ class Performance {
 
   measure(name, startMark, endMark) {
     name = `${name}`;
-    endMark = `${endMark}`;
-    startMark = startMark !== undefined ? `${startMark}` : '';
     const marks = this[kIndex][kMarks];
-    if (!marks.has(endMark) && !(endMark in nodeTiming)) {
-      const errors = lazyErrors();
-      throw new errors.ERR_INVALID_PERFORMANCE_MARK(endMark);
+    if (arguments.length >= 3) {
+      if (!marks.has(endMark) && !(endMark in nodeTiming))
+        throw new ERR_INVALID_PERFORMANCE_MARK(endMark);
+      else
+        endMark = `${endMark}`;
     }
+    startMark = startMark !== undefined ? `${startMark}` : '';
     _measure(name, startMark, endMark);
   }
 
   clearMarks(name) {
-    name = name !== undefined ? `${name}` : name;
     if (name !== undefined) {
+      name = `${name}`;
       this[kIndex][kMarks].delete(name);
       _clearMark(name);
     } else {
@@ -400,19 +425,18 @@ class Performance {
 
   timerify(fn) {
     if (typeof fn !== 'function') {
-      const errors = lazyErrors();
-      throw new errors.ERR_INVALID_ARG_TYPE('fn', 'Function', fn);
+      throw new ERR_INVALID_ARG_TYPE('fn', 'Function', fn);
     }
     if (fn[kTimerified])
       return fn[kTimerified];
     const ret = timerify(fn, fn.length);
-    Object.defineProperty(fn, kTimerified, {
+    ObjectDefineProperty(fn, kTimerified, {
       enumerable: false,
       configurable: true,
       writable: false,
       value: ret
     });
-    Object.defineProperties(ret, {
+    ObjectDefineProperties(ret, {
       [kTimerified]: {
         enumerable: false,
         configurable: true,
@@ -429,10 +453,37 @@ class Performance {
     return ret;
   }
 
+  eventLoopUtilization(util1, util2) {
+    const ls = nodeTiming.loopStart;
+
+    if (ls <= 0) {
+      return { idle: 0, active: 0, utilization: 0 };
+    }
+
+    if (util2) {
+      const idle = util1.idle - util2.idle;
+      const active = util1.active - util2.active;
+      return { idle, active, utilization: active / (idle + active) };
+    }
+
+    const idle = nodeTiming.idleTime;
+    const active = performance.now() - ls - idle;
+
+    if (!util1) {
+      return { idle, active, utilization: active / (idle + active) };
+    }
+
+    const idle_delta = idle - util1.idle;
+    const active_delta = active - util1.active;
+    const utilization = active_delta / (idle_delta + active_delta);
+    return { idle: idle_delta, active: active_delta, utilization };
+  }
+
   [kInspect]() {
     return {
       nodeTiming: this.nodeTiming,
-      timeOrigin: this.timeOrigin
+      timeOrigin: this.timeOrigin,
+      idleTime: this.idleTime,
     };
   }
 }
@@ -448,11 +499,14 @@ function getObserversList(type) {
   return list;
 }
 
-function doNotify() {
-  this[kQueued] = false;
-  this.runInAsyncScope(this[kCallback], this, this[kBuffer], this);
-  this[kBuffer][kEntries] = [];
-  L.init(this[kBuffer][kEntries]);
+function doNotify(observer) {
+  observer[kQueued] = false;
+  observer.runInAsyncScope(observer[kCallback],
+                           observer,
+                           observer[kBuffer],
+                           observer);
+  observer[kBuffer][kEntries] = [];
+  L.init(observer[kBuffer][kEntries]);
 }
 
 // Set up the callback used to receive PerformanceObserver notifications
@@ -478,11 +532,11 @@ function observersCallback(entry) {
         observer[kQueued] = true;
         // Use setImmediate instead of nextTick to give more time
         // for multiple entries to collect.
-        setImmediate(doNotify.bind(observer));
+        setImmediate(doNotify, observer);
       }
     } else {
       // If not buffering, notify immediately
-      doNotify.call(observer);
+      doNotify(observer);
     }
     current = current._idlePrev;
   }
@@ -501,6 +555,7 @@ function mapTypes(i) {
     case 'gc': return NODE_PERFORMANCE_ENTRY_TYPE_GC;
     case 'function': return NODE_PERFORMANCE_ENTRY_TYPE_FUNCTION;
     case 'http2': return NODE_PERFORMANCE_ENTRY_TYPE_HTTP2;
+    case 'http': return NODE_PERFORMANCE_ENTRY_TYPE_HTTP;
   }
 }
 
@@ -538,12 +593,33 @@ function sortedInsert(list, entry) {
   list.splice(location, 0, entry);
 }
 
+class ELDHistogram extends Histogram {
+  enable() { return this[kHandle].enable(); }
+  disable() { return this[kHandle].disable(); }
+}
+
+function monitorEventLoopDelay(options = {}) {
+  if (typeof options !== 'object' || options === null) {
+    throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
+  }
+  const { resolution = 10 } = options;
+  if (typeof resolution !== 'number') {
+    throw new ERR_INVALID_ARG_TYPE('options.resolution',
+                                   'number', resolution);
+  }
+  if (resolution <= 0 || !NumberIsSafeInteger(resolution)) {
+    throw new ERR_INVALID_OPT_VALUE.RangeError('resolution', resolution);
+  }
+  return new ELDHistogram(new _ELDHistogram(resolution));
+}
+
 module.exports = {
   performance,
-  PerformanceObserver
+  PerformanceObserver,
+  monitorEventLoopDelay
 };
 
-Object.defineProperty(module.exports, 'constants', {
+ObjectDefineProperty(module.exports, 'constants', {
   configurable: false,
   enumerable: true,
   value: constants
