@@ -1,4 +1,4 @@
-/* eslint-disable node-core/required-modules */
+/* eslint-disable node-core/require-common-first, node-core/required-modules */
 
 'use strict';
 
@@ -8,17 +8,12 @@ const path = require('path');
 
 const runjs = path.join(__dirname, '..', '..', 'benchmark', 'run.js');
 
-function runBenchmark(name, args, env) {
-  const argv = [];
-
-  for (let i = 0; i < args.length; i++) {
-    argv.push('--set');
-    argv.push(args[i]);
-  }
+function runBenchmark(name, env) {
+  const argv = ['test'];
 
   argv.push(name);
 
-  const mergedEnv = Object.assign({}, process.env, env);
+  const mergedEnv = { ...process.env, ...env };
 
   const child = fork(runjs, argv, {
     env: mergedEnv,
