@@ -6,14 +6,17 @@ if (!common.hasCrypto)
 
 const assert = require('assert');
 const {
+  constants,
   createSign,
   createVerify,
   generateKeyPair,
   generateKeyPairSync,
   publicEncrypt,
-  privateDecrypt
+  privateDecrypt,
+  sign,
+  verify
 } = require('crypto');
-const { promisify } = require('util');
+const { inspect, promisify } = require('util');
 
 // Asserts that the size of the given key (in chars or bytes) is within 10% of
 // the expected size.
@@ -31,19 +34,34 @@ function assertApproximateSize(key, expectedSize) {
 function testEncryptDecrypt(publicKey, privateKey) {
   const message = 'Hello Node.js world!';
   const plaintext = Buffer.from(message, 'utf8');
-  const ciphertext = publicEncrypt(publicKey, plaintext);
-  const received = privateDecrypt(privateKey, ciphertext);
-  assert.strictEqual(received.toString('utf8'), message);
+  for (const key of [publicKey, privateKey]) {
+    const ciphertext = publicEncrypt(key, plaintext);
+    const received = privateDecrypt(privateKey, ciphertext);
+    assert.strictEqual(received.toString('utf8'), message);
+  }
 }
 
 // Tests that a key pair can be used for signing / verification.
 function testSignVerify(publicKey, privateKey) {
-  const message = 'Hello Node.js world!';
-  const signature = createSign('SHA256').update(message)
-                                        .sign(privateKey, 'hex');
-  const okay = createVerify('SHA256').update(message)
-                                     .verify(publicKey, signature, 'hex');
-  assert(okay);
+  const message = Buffer.from('Hello Node.js world!');
+
+  function oldSign(algo, data, key) {
+    return createSign(algo).update(data).sign(key);
+  }
+
+  function oldVerify(algo, data, key, signature) {
+    return createVerify(algo).update(data).verify(key, signature);
+  }
+
+  for (const signFn of [sign, oldSign]) {
+    const signature = signFn('SHA256', message, privateKey);
+    for (const verifyFn of [verify, oldVerify]) {
+      for (const key of [publicKey, privateKey]) {
+        const okay = verifyFn('SHA256', message, key, signature);
+        assert(okay);
+      }
+    }
+  }
 }
 
 // Constructs a regular expression for a PEM-encoded key with the given label.
@@ -65,28 +83,11 @@ const pkcs8EncExp = getRegExpForPEM('ENCRYPTED PRIVATE KEY');
 const sec1Exp = getRegExpForPEM('EC PRIVATE KEY');
 const sec1EncExp = (cipher) => getRegExpForPEM('EC PRIVATE KEY', cipher);
 
-// Since our own APIs only accept PEM, not DER, we need to convert DER to PEM
-// for testing.
-function convertDERToPEM(label, der) {
-  const base64 = der.toString('base64');
-  const lines = [];
-  let i = 0;
-  while (i < base64.length) {
-    const n = Math.min(base64.length - i, 64);
-    lines.push(base64.substr(i, n));
-    i += n;
-  }
-  const body = lines.join('\n');
-  const r = `-----BEGIN ${label}-----\n${body}\n-----END ${label}-----\n`;
-  assert(getRegExpForPEM(label).test(r));
-  return r;
-}
-
 {
   // To make the test faster, we will only test sync key generation once and
   // with a relatively small key.
   const ret = generateKeyPairSync('rsa', {
-    publicExponent: 0x10001,
+    publicExponent: 3,
     modulusLength: 512,
     publicKeyEncoding: {
       type: 'pkcs1',
@@ -113,14 +114,31 @@ function convertDERToPEM(label, der) {
 }
 
 {
+  // Test sync key generation with key objects.
+  const { publicKey, privateKey } = generateKeyPairSync('rsa', {
+    modulusLength: 512
+  });
+
+  assert.strictEqual(typeof publicKey, 'object');
+  assert.strictEqual(publicKey.type, 'public');
+  assert.strictEqual(publicKey.asymmetricKeyType, 'rsa');
+
+  assert.strictEqual(typeof privateKey, 'object');
+  assert.strictEqual(privateKey.type, 'private');
+  assert.strictEqual(privateKey.asymmetricKeyType, 'rsa');
+}
+
+{
+  const publicKeyEncoding = {
+    type: 'pkcs1',
+    format: 'der'
+  };
+
   // Test async RSA key generation.
   generateKeyPair('rsa', {
     publicExponent: 0x10001,
     modulusLength: 512,
-    publicKeyEncoding: {
-      type: 'pkcs1',
-      format: 'der'
-    },
+    publicKeyEncoding,
     privateKeyEncoding: {
       type: 'pkcs1',
       format: 'pem'
@@ -128,28 +146,23 @@ function convertDERToPEM(label, der) {
   }, common.mustCall((err, publicKeyDER, privateKey) => {
     assert.ifError(err);
 
-    // The public key is encoded as DER (which is binary) instead of PEM. We
-    // will still need to convert it to PEM for testing.
     assert(Buffer.isBuffer(publicKeyDER));
-    const publicKey = convertDERToPEM('RSA PUBLIC KEY', publicKeyDER);
-    assertApproximateSize(publicKey, 180);
+    assertApproximateSize(publicKeyDER, 74);
 
     assert.strictEqual(typeof privateKey, 'string');
     assert(pkcs1PrivExp.test(privateKey));
     assertApproximateSize(privateKey, 512);
 
+    const publicKey = { key: publicKeyDER, ...publicKeyEncoding };
     testEncryptDecrypt(publicKey, privateKey);
     testSignVerify(publicKey, privateKey);
   }));
 
   // Now do the same with an encrypted private key.
   generateKeyPair('rsa', {
-    publicExponent: 0x10001,
+    publicExponent: 0x1001,
     modulusLength: 512,
-    publicKeyEncoding: {
-      type: 'pkcs1',
-      format: 'der'
-    },
+    publicKeyEncoding,
     privateKeyEncoding: {
       type: 'pkcs1',
       format: 'pem',
@@ -159,27 +172,142 @@ function convertDERToPEM(label, der) {
   }, common.mustCall((err, publicKeyDER, privateKey) => {
     assert.ifError(err);
 
-    // The public key is encoded as DER (which is binary) instead of PEM. We
-    // will still need to convert it to PEM for testing.
     assert(Buffer.isBuffer(publicKeyDER));
-    const publicKey = convertDERToPEM('RSA PUBLIC KEY', publicKeyDER);
-    assertApproximateSize(publicKey, 180);
+    assertApproximateSize(publicKeyDER, 74);
 
     assert.strictEqual(typeof privateKey, 'string');
     assert(pkcs1EncExp('AES-256-CBC').test(privateKey));
 
     // Since the private key is encrypted, signing shouldn't work anymore.
-    assert.throws(() => {
-      testSignVerify(publicKey, privateKey);
-    }, /bad decrypt|asn1 encoding routines/);
+    const publicKey = { key: publicKeyDER, ...publicKeyEncoding };
+    assert.throws(() => testSignVerify(publicKey, privateKey), {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
 
     const key = { key: privateKey, passphrase: 'secret' };
     testEncryptDecrypt(publicKey, key);
     testSignVerify(publicKey, key);
   }));
+
+  // Now do the same with an encrypted private key, but encoded as DER.
+  generateKeyPair('rsa', {
+    publicExponent: 0x10001,
+    modulusLength: 512,
+    publicKeyEncoding,
+    privateKeyEncoding: {
+      type: 'pkcs8',
+      format: 'der',
+      cipher: 'aes-256-cbc',
+      passphrase: 'secret'
+    }
+  }, common.mustCall((err, publicKeyDER, privateKeyDER) => {
+    assert.ifError(err);
+
+    assert(Buffer.isBuffer(publicKeyDER));
+    assertApproximateSize(publicKeyDER, 74);
+
+    assert(Buffer.isBuffer(privateKeyDER));
+
+    // Since the private key is encrypted, signing shouldn't work anymore.
+    const publicKey = { key: publicKeyDER, ...publicKeyEncoding };
+    assert.throws(() => {
+      testSignVerify(publicKey, {
+        key: privateKeyDER,
+        format: 'der',
+        type: 'pkcs8'
+      });
+    }, {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
+
+    // Signing should work with the correct password.
+
+    const privateKey = {
+      key: privateKeyDER,
+      format: 'der',
+      type: 'pkcs8',
+      passphrase: 'secret'
+    };
+    testEncryptDecrypt(publicKey, privateKey);
+    testSignVerify(publicKey, privateKey);
+  }));
+
+  // Now do the same with an encrypted private key, but encoded as DER.
+  generateKeyPair('rsa', {
+    publicExponent: 0x10001,
+    modulusLength: 512,
+    publicKeyEncoding,
+    privateKeyEncoding: {
+      type: 'pkcs8',
+      format: 'der'
+    }
+  }, common.mustCall((err, publicKeyDER, privateKeyDER) => {
+    assert.ifError(err);
+
+    assert(Buffer.isBuffer(publicKeyDER));
+    assertApproximateSize(publicKeyDER, 74);
+
+    assert(Buffer.isBuffer(privateKeyDER));
+
+    const publicKey = { key: publicKeyDER, ...publicKeyEncoding };
+    const privateKey = {
+      key: privateKeyDER,
+      format: 'der',
+      type: 'pkcs8',
+      passphrase: 'secret'
+    };
+    testEncryptDecrypt(publicKey, privateKey);
+    testSignVerify(publicKey, privateKey);
+  }));
 }
 
 {
+  // Test RSA-PSS.
+  generateKeyPair('rsa-pss', {
+    modulusLength: 512,
+    saltLength: 16,
+    hash: 'sha256',
+    mgf1Hash: 'sha256'
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    assert.strictEqual(publicKey.type, 'public');
+    assert.strictEqual(publicKey.asymmetricKeyType, 'rsa-pss');
+
+    assert.strictEqual(privateKey.type, 'private');
+    assert.strictEqual(privateKey.asymmetricKeyType, 'rsa-pss');
+
+    // Unlike RSA, RSA-PSS does not allow encryption.
+    assert.throws(() => {
+      testEncryptDecrypt(publicKey, privateKey);
+    }, /operation not supported for this keytype/);
+
+    // RSA-PSS also does not permit signing with PKCS1 padding.
+    assert.throws(() => {
+      testSignVerify({
+        key: publicKey,
+        padding: constants.RSA_PKCS1_PADDING
+      }, {
+        key: privateKey,
+        padding: constants.RSA_PKCS1_PADDING
+      });
+    }, /illegal or unsupported padding mode/);
+
+    // The padding should correctly default to RSA_PKCS1_PSS_PADDING now.
+    testSignVerify(publicKey, privateKey);
+  }));
+}
+
+{
+  const privateKeyEncoding = {
+    type: 'pkcs8',
+    format: 'der'
+  };
+
   // Test async DSA key generation.
   generateKeyPair('dsa', {
     modulusLength: 512,
@@ -189,10 +317,9 @@ function convertDERToPEM(label, der) {
       format: 'pem'
     },
     privateKeyEncoding: {
-      type: 'pkcs8',
-      format: 'der',
       cipher: 'aes-128-cbc',
-      passphrase: 'secret'
+      passphrase: 'secret',
+      ...privateKeyEncoding
     }
   }, common.mustCall((err, publicKey, privateKeyDER) => {
     assert.ifError(err);
@@ -201,19 +328,26 @@ function convertDERToPEM(label, der) {
     assert(spkiExp.test(publicKey));
     // The private key is DER-encoded.
     assert(Buffer.isBuffer(privateKeyDER));
-    const privateKey = convertDERToPEM('ENCRYPTED PRIVATE KEY', privateKeyDER);
 
     assertApproximateSize(publicKey, 440);
-    assertApproximateSize(privateKey, 512);
+    assertApproximateSize(privateKeyDER, 336);
 
     // Since the private key is encrypted, signing shouldn't work anymore.
     assert.throws(() => {
-      testSignVerify(publicKey, privateKey);
-    }, /bad decrypt|asn1 encoding routines/);
+      return testSignVerify(publicKey, {
+        key: privateKeyDER,
+        ...privateKeyEncoding
+      });
+    }, {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
 
     // Signing should work with the correct password.
     testSignVerify(publicKey, {
-      key: privateKey,
+      key: privateKeyDER,
+      ...privateKeyEncoding,
       passphrase: 'secret'
     });
   }));
@@ -244,6 +378,30 @@ function convertDERToPEM(label, der) {
     testSignVerify(publicKey, privateKey);
   }));
 
+  // Test async elliptic curve key generation, e.g. for ECDSA, with a SEC1
+  // private key with paramEncoding explicit.
+  generateKeyPair('ec', {
+    namedCurve: 'prime256v1',
+    paramEncoding: 'explicit',
+    publicKeyEncoding: {
+      type: 'spki',
+      format: 'pem'
+    },
+    privateKeyEncoding: {
+      type: 'sec1',
+      format: 'pem'
+    }
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    assert.strictEqual(typeof publicKey, 'string');
+    assert(spkiExp.test(publicKey));
+    assert.strictEqual(typeof privateKey, 'string');
+    assert(sec1Exp.test(privateKey));
+
+    testSignVerify(publicKey, privateKey);
+  }));
+
   // Do the same with an encrypted private key.
   generateKeyPair('ec', {
     namedCurve: 'prime256v1',
@@ -267,9 +425,43 @@ function convertDERToPEM(label, der) {
     assert(sec1EncExp('AES-128-CBC').test(privateKey));
 
     // Since the private key is encrypted, signing shouldn't work anymore.
-    assert.throws(() => {
-      testSignVerify(publicKey, privateKey);
-    }, /bad decrypt|asn1 encoding routines/);
+    assert.throws(() => testSignVerify(publicKey, privateKey), {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
+
+    testSignVerify(publicKey, { key: privateKey, passphrase: 'secret' });
+  }));
+
+  // Do the same with an encrypted private key with paramEncoding explicit.
+  generateKeyPair('ec', {
+    namedCurve: 'prime256v1',
+    paramEncoding: 'explicit',
+    publicKeyEncoding: {
+      type: 'spki',
+      format: 'pem'
+    },
+    privateKeyEncoding: {
+      type: 'sec1',
+      format: 'pem',
+      cipher: 'aes-128-cbc',
+      passphrase: 'secret'
+    }
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    assert.strictEqual(typeof publicKey, 'string');
+    assert(spkiExp.test(publicKey));
+    assert.strictEqual(typeof privateKey, 'string');
+    assert(sec1EncExp('AES-128-CBC').test(privateKey));
+
+    // Since the private key is encrypted, signing shouldn't work anymore.
+    assert.throws(() => testSignVerify(publicKey, privateKey), {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
 
     testSignVerify(publicKey, { key: privateKey, passphrase: 'secret' });
   }));
@@ -279,7 +471,7 @@ function convertDERToPEM(label, der) {
   // Test async elliptic curve key generation, e.g. for ECDSA, with an encrypted
   // private key.
   generateKeyPair('ec', {
-    namedCurve: 'P-192',
+    namedCurve: 'P-256',
     paramEncoding: 'named',
     publicKeyEncoding: {
       type: 'spki',
@@ -300,15 +492,76 @@ function convertDERToPEM(label, der) {
     assert(pkcs8EncExp.test(privateKey));
 
     // Since the private key is encrypted, signing shouldn't work anymore.
-    assert.throws(() => {
-      testSignVerify(publicKey, privateKey);
-    }, /bad decrypt|asn1 encoding routines/);
+    assert.throws(() => testSignVerify(publicKey, privateKey), {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
 
     testSignVerify(publicKey, {
       key: privateKey,
       passphrase: 'top secret'
     });
   }));
+
+  // Test async elliptic curve key generation, e.g. for ECDSA, with an encrypted
+  // private key with paramEncoding explicit.
+  generateKeyPair('ec', {
+    namedCurve: 'P-256',
+    paramEncoding: 'explicit',
+    publicKeyEncoding: {
+      type: 'spki',
+      format: 'pem'
+    },
+    privateKeyEncoding: {
+      type: 'pkcs8',
+      format: 'pem',
+      cipher: 'aes-128-cbc',
+      passphrase: 'top secret'
+    }
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    assert.strictEqual(typeof publicKey, 'string');
+    assert(spkiExp.test(publicKey));
+    assert.strictEqual(typeof privateKey, 'string');
+    assert(pkcs8EncExp.test(privateKey));
+
+    // Since the private key is encrypted, signing shouldn't work anymore.
+    assert.throws(() => testSignVerify(publicKey, privateKey), {
+      name: 'TypeError',
+      code: 'ERR_MISSING_PASSPHRASE',
+      message: 'Passphrase required for encrypted key'
+    });
+
+    testSignVerify(publicKey, {
+      key: privateKey,
+      passphrase: 'top secret'
+    });
+  }));
+}
+
+// Test invalid parameter encoding.
+{
+  assert.throws(() => generateKeyPairSync('ec', {
+    namedCurve: 'P-256',
+    paramEncoding: 'otherEncoding',
+    publicKeyEncoding: {
+      type: 'spki',
+      format: 'pem'
+    },
+    privateKeyEncoding: {
+      type: 'pkcs8',
+      format: 'pem',
+      cipher: 'aes-128-cbc',
+      passphrase: 'top secret'
+    }
+  }), {
+    name: 'TypeError',
+    code: 'ERR_INVALID_OPT_VALUE',
+    message: 'The value "otherEncoding" is invalid for ' +
+    'option "paramEncoding"'
+  });
 }
 
 {
@@ -342,36 +595,88 @@ function convertDERToPEM(label, der) {
 {
   // Test invalid key types.
   for (const type of [undefined, null, 0]) {
-    common.expectsError(() => generateKeyPairSync(type, {}), {
-      type: TypeError,
+    assert.throws(() => generateKeyPairSync(type, {}), {
+      name: 'TypeError',
       code: 'ERR_INVALID_ARG_TYPE',
-      message: 'The "type" argument must be of type string. Received type ' +
-               typeof type
+      message: 'The "type" argument must be of type string.' +
+               common.invalidArgTypeHelper(type)
     });
   }
 
-  common.expectsError(() => generateKeyPairSync('rsa2', {}), {
-    type: TypeError,
+  assert.throws(() => generateKeyPairSync('rsa2', {}), {
+    name: 'TypeError',
     code: 'ERR_INVALID_ARG_VALUE',
-    message: "The argument 'type' must be one of " +
-             "'rsa', 'dsa', 'ec'. Received 'rsa2'"
+    message: "The argument 'type' must be a supported key type. Received 'rsa2'"
   });
 }
 
 {
   // Test keygen without options object.
-  common.expectsError(() => generateKeyPair('rsa', common.mustNotCall()), {
-    type: TypeError,
+  assert.throws(() => generateKeyPair('rsa', common.mustNotCall()), {
+    name: 'TypeError',
     code: 'ERR_INVALID_ARG_TYPE',
-    message: 'The "options" argument must be of ' +
-      'type object. Received type undefined'
+    message: 'The "options" argument must be of type object. ' +
+      'Received undefined'
+  });
+
+  // Even if no options are required, it should be impossible to pass anything
+  // but an object (or undefined).
+  assert.throws(() => generateKeyPair('ed448', 0, common.mustNotCall()), {
+    name: 'TypeError',
+    code: 'ERR_INVALID_ARG_TYPE',
+    message: 'The "options" argument must be of type object. ' +
+      'Received type number (0)'
   });
 }
 
 {
-  // Missing / invalid publicKeyEncoding.
-  for (const enc of [undefined, null, 0, 'a', true]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+  // If no publicKeyEncoding is specified, a key object should be returned.
+  generateKeyPair('rsa', {
+    modulusLength: 1024,
+    privateKeyEncoding: {
+      type: 'pkcs1',
+      format: 'pem'
+    }
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    assert.strictEqual(typeof publicKey, 'object');
+    assert.strictEqual(publicKey.type, 'public');
+    assert.strictEqual(publicKey.asymmetricKeyType, 'rsa');
+
+    // The private key should still be a string.
+    assert.strictEqual(typeof privateKey, 'string');
+
+    testEncryptDecrypt(publicKey, privateKey);
+    testSignVerify(publicKey, privateKey);
+  }));
+
+  // If no privateKeyEncoding is specified, a key object should be returned.
+  generateKeyPair('rsa', {
+    modulusLength: 1024,
+    publicKeyEncoding: {
+      type: 'pkcs1',
+      format: 'pem'
+    }
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    // The public key should still be a string.
+    assert.strictEqual(typeof publicKey, 'string');
+
+    assert.strictEqual(typeof privateKey, 'object');
+    assert.strictEqual(privateKey.type, 'private');
+    assert.strictEqual(privateKey.asymmetricKeyType, 'rsa');
+
+    testEncryptDecrypt(publicKey, privateKey);
+    testSignVerify(publicKey, privateKey);
+  }));
+}
+
+{
+  // Invalid publicKeyEncoding.
+  for (const enc of [0, 'a', true]) {
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: enc,
       privateKeyEncoding: {
@@ -379,7 +684,7 @@ function convertDERToPEM(label, der) {
         format: 'pem'
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
       message: `The value "${enc}" is invalid for option "publicKeyEncoding"`
     });
@@ -387,7 +692,7 @@ function convertDERToPEM(label, der) {
 
   // Missing publicKeyEncoding.type.
   for (const type of [undefined, null, 0, true, {}]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type,
@@ -398,16 +703,17 @@ function convertDERToPEM(label, der) {
         format: 'pem'
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${type}" is invalid for option ` +
+      message: `The value "${inspect(type)}" is invalid for option ` +
                '"publicKeyEncoding.type"'
     });
   }
 
   // Missing / invalid publicKeyEncoding.format.
   for (const format of [undefined, null, 0, false, 'a', {}]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+    const expected = typeof format === 'string' ? format : inspect(format);
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type: 'pkcs1',
@@ -418,16 +724,16 @@ function convertDERToPEM(label, der) {
         format: 'pem'
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${format}" is invalid for option ` +
+      message: `The value "${expected}" is invalid for option ` +
                '"publicKeyEncoding.format"'
     });
   }
 
-  // Missing / invalid privateKeyEncoding.
-  for (const enc of [undefined, null, 0, 'a', true]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+  // Invalid privateKeyEncoding.
+  for (const enc of [0, 'a', true]) {
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type: 'pkcs1',
@@ -435,7 +741,7 @@ function convertDERToPEM(label, der) {
       },
       privateKeyEncoding: enc
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
       message: `The value "${enc}" is invalid for option "privateKeyEncoding"`
     });
@@ -443,7 +749,7 @@ function convertDERToPEM(label, der) {
 
   // Missing / invalid privateKeyEncoding.type.
   for (const type of [undefined, null, 0, true, {}]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type: 'pkcs1',
@@ -454,16 +760,17 @@ function convertDERToPEM(label, der) {
         format: 'pem'
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${type}" is invalid for option ` +
+      message: `The value "${inspect(type)}" is invalid for option ` +
                '"privateKeyEncoding.type"'
     });
   }
 
   // Missing / invalid privateKeyEncoding.format.
   for (const format of [undefined, null, 0, false, 'a', {}]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+    const expected = typeof format === 'string' ? format : inspect(format);
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type: 'pkcs1',
@@ -474,16 +781,16 @@ function convertDERToPEM(label, der) {
         format
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${format}" is invalid for option ` +
+      message: `The value "${expected}" is invalid for option ` +
                '"privateKeyEncoding.format"'
     });
   }
 
-  // cipher of invalid type.
+  // Cipher of invalid type.
   for (const cipher of [0, true, {}]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type: 'pkcs1',
@@ -495,15 +802,15 @@ function convertDERToPEM(label, der) {
         cipher
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${cipher}" is invalid for option ` +
+      message: `The value "${inspect(cipher)}" is invalid for option ` +
                '"privateKeyEncoding.cipher"'
     });
   }
 
   // Invalid cipher.
-  common.expectsError(() => generateKeyPairSync('rsa', {
+  assert.throws(() => generateKeyPairSync('rsa', {
     modulusLength: 4096,
     publicKeyEncoding: {
       type: 'pkcs1',
@@ -516,13 +823,14 @@ function convertDERToPEM(label, der) {
       passphrase: 'secret'
     }
   }), {
-    type: Error,
+    name: 'Error',
+    code: 'ERR_CRYPTO_UNKNOWN_CIPHER',
     message: 'Unknown cipher'
   });
 
-  // cipher, but no valid passphrase.
+  // Cipher, but no valid passphrase.
   for (const passphrase of [undefined, null, 5, false, true]) {
-    common.expectsError(() => generateKeyPairSync('rsa', {
+    assert.throws(() => generateKeyPairSync('rsa', {
       modulusLength: 4096,
       publicKeyEncoding: {
         type: 'pkcs1',
@@ -535,7 +843,7 @@ function convertDERToPEM(label, der) {
         passphrase
       }
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
       message: `The value "${passphrase}" is invalid for option ` +
                '"privateKeyEncoding.passphrase"'
@@ -544,12 +852,12 @@ function convertDERToPEM(label, der) {
 
   // Test invalid callbacks.
   for (const cb of [undefined, null, 0, {}]) {
-    common.expectsError(() => generateKeyPair('rsa', {
+    assert.throws(() => generateKeyPair('rsa', {
       modulusLength: 512,
       publicKeyEncoding: { type: 'pkcs1', format: 'pem' },
       privateKeyEncoding: { type: 'pkcs1', format: 'pem' }
     }, cb), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_CALLBACK'
     });
   }
@@ -559,25 +867,29 @@ function convertDERToPEM(label, der) {
 {
   // Test invalid modulus lengths.
   for (const modulusLength of [undefined, null, 'a', true, {}, [], 512.1, -1]) {
-    common.expectsError(() => generateKeyPair('rsa', {
+    const expected = typeof modulusLength === 'string' ?
+      modulusLength : inspect(modulusLength);
+    assert.throws(() => generateKeyPair('rsa', {
       modulusLength
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${modulusLength}" is invalid for option ` +
+      message: `The value "${expected}" is invalid for option ` +
                '"modulusLength"'
     });
   }
 
   // Test invalid exponents.
   for (const publicExponent of ['a', true, {}, [], 3.5, -1]) {
-    common.expectsError(() => generateKeyPair('rsa', {
+    const expected = typeof publicExponent === 'string' ?
+      publicExponent : inspect(publicExponent);
+    assert.throws(() => generateKeyPair('rsa', {
       modulusLength: 4096,
       publicExponent
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${publicExponent}" is invalid for option ` +
+      message: `The value "${expected}" is invalid for option ` +
                '"publicExponent"'
     });
   }
@@ -587,25 +899,29 @@ function convertDERToPEM(label, der) {
 {
   // Test invalid modulus lengths.
   for (const modulusLength of [undefined, null, 'a', true, {}, [], 4096.1]) {
-    common.expectsError(() => generateKeyPair('dsa', {
+    const expected = typeof modulusLength === 'string' ?
+      modulusLength : inspect(modulusLength);
+    assert.throws(() => generateKeyPair('dsa', {
       modulusLength
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${modulusLength}" is invalid for option ` +
+      message: `The value "${expected}" is invalid for option ` +
                '"modulusLength"'
     });
   }
 
   // Test invalid divisor lengths.
   for (const divisorLength of ['a', true, {}, [], 4096.1]) {
-    common.expectsError(() => generateKeyPair('dsa', {
+    const expected = typeof divisorLength === 'string' ?
+      divisorLength : inspect(divisorLength);
+    assert.throws(() => generateKeyPair('dsa', {
       modulusLength: 2048,
       divisorLength
     }), {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${divisorLength}" is invalid for option ` +
+      message: `The value "${expected}" is invalid for option ` +
                '"divisorLength"'
     });
   }
@@ -614,36 +930,36 @@ function convertDERToPEM(label, der) {
 // Test EC parameters.
 {
   // Test invalid curves.
-  common.expectsError(() => {
+  assert.throws(() => {
     generateKeyPairSync('ec', {
       namedCurve: 'abcdef',
       publicKeyEncoding: { type: 'spki', format: 'pem' },
       privateKeyEncoding: { type: 'sec1', format: 'pem' }
     });
   }, {
-    type: TypeError,
+    name: 'TypeError',
     message: 'Invalid ECDH curve name'
   });
 
   // Test error type when curve is not a string
   for (const namedCurve of [true, {}, [], 123]) {
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync('ec', {
         namedCurve,
         publicKeyEncoding: { type: 'spki', format: 'pem' },
         privateKeyEncoding: { type: 'sec1', format: 'pem' }
       });
     }, {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
-      message: `The value "${namedCurve}" is invalid for option ` +
+      message: `The value "${inspect(namedCurve)}" is invalid for option ` +
                '"namedCurve"'
     });
   }
 
   // It should recognize both NIST and standard curve names.
   generateKeyPair('ec', {
-    namedCurve: 'P-192',
+    namedCurve: 'P-256',
     publicKeyEncoding: { type: 'spki', format: 'pem' },
     privateKeyEncoding: { type: 'pkcs8', format: 'pem' }
   }, common.mustCall((err, publicKey, privateKey) => {
@@ -651,7 +967,7 @@ function convertDERToPEM(label, der) {
   }));
 
   generateKeyPair('ec', {
-    namedCurve: 'secp192k1',
+    namedCurve: 'secp256k1',
     publicKeyEncoding: { type: 'spki', format: 'pem' },
     privateKeyEncoding: { type: 'pkcs8', format: 'pem' }
   }, common.mustCall((err, publicKey, privateKey) => {
@@ -659,34 +975,134 @@ function convertDERToPEM(label, der) {
   }));
 }
 
+// Test EdDSA key generation.
+{
+  if (!/^1\.1\.0/.test(process.versions.openssl)) {
+    ['ed25519', 'ed448', 'x25519', 'x448'].forEach((keyType) => {
+      generateKeyPair(keyType, common.mustCall((err, publicKey, privateKey) => {
+        assert.ifError(err);
+
+        assert.strictEqual(publicKey.type, 'public');
+        assert.strictEqual(publicKey.asymmetricKeyType, keyType);
+
+        assert.strictEqual(privateKey.type, 'private');
+        assert.strictEqual(privateKey.asymmetricKeyType, keyType);
+      }));
+    });
+  }
+}
+
+// Test classic Diffie-Hellman key generation.
+{
+  generateKeyPair('dh', {
+    primeLength: 1024
+  }, common.mustCall((err, publicKey, privateKey) => {
+    assert.ifError(err);
+
+    assert.strictEqual(publicKey.type, 'public');
+    assert.strictEqual(publicKey.asymmetricKeyType, 'dh');
+
+    assert.strictEqual(privateKey.type, 'private');
+    assert.strictEqual(privateKey.asymmetricKeyType, 'dh');
+  }));
+
+  assert.throws(() => {
+    generateKeyPair('dh', common.mustNotCall());
+  }, {
+    name: 'TypeError',
+    code: 'ERR_INVALID_ARG_TYPE',
+    message: 'The "options" argument must be of type object. Received undefined'
+  });
+
+  assert.throws(() => {
+    generateKeyPair('dh', {}, common.mustNotCall());
+  }, {
+    name: 'TypeError',
+    code: 'ERR_MISSING_OPTION',
+    message: 'At least one of the group, prime, or primeLength options is ' +
+             'required'
+  });
+
+  assert.throws(() => {
+    generateKeyPair('dh', {
+      group: 'modp0'
+    }, common.mustNotCall());
+  }, {
+    name: 'Error',
+    code: 'ERR_CRYPTO_UNKNOWN_DH_GROUP',
+    message: 'Unknown DH group'
+  });
+
+  // Test incompatible options.
+  const allOpts = {
+    group: 'modp5',
+    prime: Buffer.alloc(0),
+    primeLength: 1024,
+    generator: 2
+  };
+  const incompatible = [
+    ['group', 'prime'],
+    ['group', 'primeLength'],
+    ['group', 'generator'],
+    ['prime', 'primeLength']
+  ];
+  for (const [opt1, opt2] of incompatible) {
+    assert.throws(() => {
+      generateKeyPairSync('dh', {
+        [opt1]: allOpts[opt1],
+        [opt2]: allOpts[opt2]
+      });
+    }, {
+      name: 'TypeError',
+      code: 'ERR_INCOMPATIBLE_OPTION_PAIR',
+      message: `Option "${opt1}" cannot be used in combination with option ` +
+               `"${opt2}"`
+    });
+  }
+}
+
 // Test invalid key encoding types.
 {
   // Invalid public key type.
   for (const type of ['foo', 'pkcs8', 'sec1']) {
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync('rsa', {
         modulusLength: 4096,
         publicKeyEncoding: { type, format: 'pem' },
         privateKeyEncoding: { type: 'pkcs8', format: 'pem' }
       });
     }, {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
       message: `The value "${type}" is invalid for option ` +
                '"publicKeyEncoding.type"'
     });
   }
 
+  // Invalid hash value.
+  for (const hashValue of [123, true, {}, []]) {
+    assert.throws(() => {
+      generateKeyPairSync('rsa-pss', {
+        modulusLength: 4096,
+        hash: hashValue
+      });
+    }, {
+      name: 'TypeError',
+      code: 'ERR_INVALID_OPT_VALUE',
+      message: `The value "${inspect(hashValue)}" is invalid for option "hash"`
+    });
+  }
+
   // Invalid private key type.
   for (const type of ['foo', 'spki']) {
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync('rsa', {
         modulusLength: 4096,
         publicKeyEncoding: { type: 'spki', format: 'pem' },
         privateKeyEncoding: { type, format: 'pem' }
       });
     }, {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE',
       message: `The value "${type}" is invalid for option ` +
                '"privateKeyEncoding.type"'
@@ -695,7 +1111,7 @@ function convertDERToPEM(label, der) {
 
   // Key encoding doesn't match key type.
   for (const type of ['dsa', 'ec']) {
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync(type, {
         modulusLength: 4096,
         namedCurve: 'P-256',
@@ -703,12 +1119,12 @@ function convertDERToPEM(label, der) {
         privateKeyEncoding: { type: 'pkcs8', format: 'pem' }
       });
     }, {
-      type: Error,
+      name: 'Error',
       code: 'ERR_CRYPTO_INCOMPATIBLE_KEY_OPTIONS',
       message: 'The selected key encoding pkcs1 can only be used for RSA keys.'
     });
 
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync(type, {
         modulusLength: 4096,
         namedCurve: 'P-256',
@@ -716,21 +1132,21 @@ function convertDERToPEM(label, der) {
         privateKeyEncoding: { type: 'pkcs1', format: 'pem' }
       });
     }, {
-      type: Error,
+      name: 'Error',
       code: 'ERR_CRYPTO_INCOMPATIBLE_KEY_OPTIONS',
       message: 'The selected key encoding pkcs1 can only be used for RSA keys.'
     });
   }
 
   for (const type of ['rsa', 'dsa']) {
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync(type, {
         modulusLength: 4096,
         publicKeyEncoding: { type: 'spki', format: 'pem' },
         privateKeyEncoding: { type: 'sec1', format: 'pem' }
       });
     }, {
-      type: Error,
+      name: 'Error',
       code: 'ERR_CRYPTO_INCOMPATIBLE_KEY_OPTIONS',
       message: 'The selected key encoding sec1 can only be used for EC keys.'
     });
@@ -738,7 +1154,7 @@ function convertDERToPEM(label, der) {
 
   // Attempting to encrypt a DER-encoded, non-PKCS#8 key.
   for (const type of ['pkcs1', 'sec1']) {
-    common.expectsError(() => {
+    assert.throws(() => {
       generateKeyPairSync(type === 'pkcs1' ? 'rsa' : 'ec', {
         modulusLength: 4096,
         namedCurve: 'P-256',
@@ -751,9 +1167,46 @@ function convertDERToPEM(label, der) {
         }
       });
     }, {
-      type: Error,
+      name: 'Error',
       code: 'ERR_CRYPTO_INCOMPATIBLE_KEY_OPTIONS',
       message: `The selected key encoding ${type} does not support encryption.`
     });
   }
 }
+{
+  // Test RSA-PSS.
+  assert.throws(
+    () => {
+      generateKeyPair('rsa-pss', {
+        modulusLength: 512,
+        saltLength: 16,
+        hash: 'sha256',
+        mgf1Hash: undefined
+      });
+    },
+    {
+      name: 'TypeError',
+      code: 'ERR_INVALID_CALLBACK',
+      message: 'Callback must be a function. Received undefined'
+    }
+  );
+
+  for (const mgf1Hash of [null, 0, false, {}, []]) {
+    const expected = inspect(mgf1Hash);
+    assert.throws(
+      () => {
+        generateKeyPair('rsa-pss', {
+          modulusLength: 512,
+          saltLength: 16,
+          hash: 'sha256',
+          mgf1Hash
+        });
+      },
+      {
+        name: 'TypeError',
+        code: 'ERR_INVALID_OPT_VALUE',
+        message: `The value "${expected}" is invalid for option "mgf1Hash"`
+      }
+    );
+  }
+}
