@@ -13,7 +13,7 @@ child.stdout.on('data', (data) => {
 });
 
 child.on('exit', common.mustCall(() => {
-  const results = output.replace(/^> /mg, '').split('\n');
+  const results = output.replace(/^> /mg, '').split('\n').slice(2);
   assert.deepStrictEqual(results, ['undefined', 'true', 'true', '']);
 }));
 
