@@ -18,33 +18,47 @@ assert.strictEqual(
   false
 );
 
-common.expectsError(
+{
+  // Test TypedArrays with different lengths but equal byteLengths.
+  const buf = crypto.randomBytes(16).buffer;
+  const a1 = new Uint8Array(buf);
+  const a2 = new Uint16Array(buf);
+  const a3 = new Uint32Array(buf);
+
+  for (const left of [a1, a2, a3]) {
+    for (const right of [a1, a2, a3]) {
+      assert.strictEqual(crypto.timingSafeEqual(left, right), true);
+    }
+  }
+}
+
+assert.throws(
   () => crypto.timingSafeEqual(Buffer.from([1, 2, 3]), Buffer.from([1, 2])),
   {
     code: 'ERR_CRYPTO_TIMING_SAFE_EQUAL_LENGTH',
-    type: RangeError,
-    message: 'Input buffers must have the same length'
+    name: 'RangeError',
+    message: 'Input buffers must have the same byte length'
   }
 );
 
-common.expectsError(
+assert.throws(
   () => crypto.timingSafeEqual('not a buffer', Buffer.from([1, 2])),
   {
     code: 'ERR_INVALID_ARG_TYPE',
-    type: TypeError,
+    name: 'TypeError',
     message:
-      'The "buf1" argument must be one of type Buffer, TypedArray, or ' +
-      'DataView. Received type string'
+      'The "buf1" argument must be an instance of Buffer, TypedArray, or ' +
+      'DataView.'
   }
 );
 
-common.expectsError(
+assert.throws(
   () => crypto.timingSafeEqual(Buffer.from([1, 2]), 'not a buffer'),
   {
     code: 'ERR_INVALID_ARG_TYPE',
-    type: TypeError,
+    name: 'TypeError',
     message:
-      'The "buf2" argument must be one of type Buffer, TypedArray, or ' +
-      'DataView. Received type string'
+      'The "buf2" argument must be an instance of Buffer, TypedArray, or ' +
+      'DataView.'
   }
 );
